import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xlis import XLISExchangeCalendar
from .test_exchange_calendar import EuronextCalendarTestBase


class TestXLISCalendar(EuronextCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XLISExchangeCalendar

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=13, minutes=5)

    @pytest.fixture
    def non_early_closes_sample_time(self):
        yield pd.Timedelta(hours=16, minutes=30)

    @pytest.fixture
    def additional_regular_holidays_sample(self):
        yield [
            # Final observance of these regular holidays
            "2002-02-12",  # Carnival
            "2002-05-30",  # Corpus Christi Day
            "2002-04-25",  # Liberty Day
            "2002-06-10",  # Portugal Day
            "2001-06-13",  # Saint Anthony's Day
            "2002-08-15",  # Assumption Day
            "2001-10-05",  # Republic Day
            "2002-11-01",  # All Saints Day
            "2000-12-01",  # Independence Day
            "2000-12-08",  # Immaculate Conception
            "2002-12-24",  # Christmas Eve
        ]

    @pytest.fixture
    def additional_non_holidays_sample(self):
        yield [
            # First year when previous regular holiday no longer observed
            "2003-03-04",  # Carnival
            "2003-06-16",  # Corpus Christi Day
            "2003-04-25",  # Liberty Day
            "2003-06-10",  # Portugal Day
            "2002-06-13",  # Saint Anthony's Day
            "2003-08-15",  # Assumption Day
            "2004-10-05",  # Republic Day
            "2004-11-01",  # All Saints Day
            "2003-12-01",  # Independence Day
            "2003-12-08",  # Immaculate Conception
            "2003-12-24",  # Christmas Eve
        ]



"""Tests for calendar_helpers module."""

from collections import abc
import datetime
from datetime import time
import itertools
import operator
import re
from zoneinfo import ZoneInfo

import numpy as np
import pandas as pd
import pytest
from hypothesis import assume, given, settings, HealthCheck
from hypothesis import strategies as st
from pandas.testing import assert_index_equal

from exchange_calendars import ExchangeCalendar
from exchange_calendars import calendar_helpers as m
from exchange_calendars import calendar_utils, errors
from exchange_calendars.calendar_helpers import UTC
from exchange_calendars.calendar_utils import XTAEExchangeCalendar

from .test_exchange_calendar import Answers

# TODO tests for next_divider_idx, previous_divider_idx, compute_minutes (#15)


def test_constants():
    # Just to make sure they aren't inadvertently changed
    assert m.UTC is ZoneInfo("UTC")
    assert m.NANOSECONDS_PER_MINUTE == int(6e10)
    assert m.NP_NAT == pd.NaT.value


@pytest.fixture(scope="class")
def one_min() -> abc.Iterator[pd.Timedelta]:
    yield pd.Timedelta(1, "min")


def test_is_date(one_min):
    f = m.is_date
    T = pd.Timestamp

    assert f(T("2021-11-02"))
    assert f(T("2021-11-02 00:00"))
    assert f(T("2021-11-02 00:00:00.0000000"))
    assert not f(T("2021-11-02 00:00:00.000001"))
    assert not f(T("2021-11-01 23:59:00.999999"))
    assert not f(T("2021-11-02 12:00"))

    tz = ZoneInfo("America/New_York")
    minutes = [
        T("2021-11-02", tz=UTC),
        T("2021-11-02", tz=tz),
        T("2021-11-02", tz=UTC).tz_convert(tz),
    ]
    for minute in minutes:
        assert not f(minute)
        assert not f(minute + one_min)


def test_is_utc():
    f = m.to_utc
    T = pd.Timestamp

    expected = T("2021-11-02", tz=UTC)
    assert f(T("2021-11-02", tz=UTC)) == expected
    assert f(T("2021-11-02")) == expected

    expected = T("2021-11-02 13:33", tz=UTC)
    assert f(T("2021-11-02 13:33")) == expected
    assert f(T("2021-11-02 09:33", tz=ZoneInfo("America/New_York"))) == expected


@pytest.fixture
def one_day() -> abc.Iterator[pd.DateOffset]:
    yield pd.DateOffset(days=1)


# all fixtures with respect to XHKG
@pytest.fixture
def calendar() -> abc.Iterator[calendar_utils.XHKGExchangeCalendar]:
    yield calendar_utils.XHKGExchangeCalendar()


@pytest.fixture
def param_name() -> abc.Iterator[str]:
    yield "parameter's_name"


@pytest.fixture(params=[True, False])
def utc(request) -> abc.Iterator[str]:
    yield request.param


@pytest.fixture(params=["2021-13-13", ("2021-06-06",), "not a timestamp"])
def malformed(request) -> abc.Iterator[str]:
    yield request.param


@pytest.fixture
def minute() -> abc.Iterator[str]:
    yield "2021-06-02 23:00"


@pytest.fixture(
    params=[
        "2021-06-02 23:00",
        pd.Timestamp("2021-06-02 23:00"),
        pd.Timestamp("2021-06-02 23:00", tz=UTC),
    ]
)
def minute_mult(request) -> abc.Iterator[str | pd.Timestamp]:
    yield request.param


@pytest.fixture
def date(calendar) -> abc.Iterator[str]:
    """Date that does not represent a session of `calendar`."""
    date_ = "2021-06-05"
    assert pd.Timestamp(date_) not in calendar.schedule.index
    yield date_


@pytest.fixture(params=["2021-06-05", pd.Timestamp("2021-06-05")])
def date_mult(request, calendar) -> abc.Iterator[str | pd.Timestamp]:
    """Date that does not represent a session of `calendar`."""
    date = request.param
    ts = pd.Timestamp(date)
    assert ts not in calendar.schedule.index
    yield date


@pytest.fixture
def second() -> abc.Iterator[str]:
    yield "2021-06-02 23:01:30"


@pytest.fixture(params=["left", "right", "both", "neither"])
def sides(request) -> abc.Iterator[str]:
    yield request.param


@pytest.fixture
def session() -> abc.Iterator[str]:
    yield "2021-06-02"


@pytest.fixture
def trading_minute() -> abc.Iterator[str]:
    yield "2021-06-02 05:30"


@pytest.fixture
def minute_too_early(calendar, one_min) -> abc.Iterator[pd.Timestamp]:
    yield calendar.first_minute - one_min


@pytest.fixture
def minute_too_late(calendar, one_min) -> abc.Iterator[pd.Timestamp]:
    yield calendar.last_minute + one_min


@pytest.fixture
def date_too_early(calendar, one_day) -> abc.Iterator[pd.Timestamp]:
    yield calendar.first_session - one_day


@pytest.fixture
def date_too_late(calendar, one_day) -> abc.Iterator[pd.Timestamp]:
    yield calendar.last_session + one_day


def test_parse_timestamp_with_date(date_mult, param_name, calendar, utc):
    date = date_mult
    date_is_utc_ts = isinstance(date, pd.Timestamp) and date.tz is not None
    dt = m.parse_timestamp(date, param_name, calendar, utc=utc)
    if not utc and not date_is_utc_ts:
        assert dt == pd.Timestamp("2021-06-05")
    else:
        assert dt == pd.Timestamp("2021-06-05", tz=UTC)
    assert dt == dt.floor("min")


def test_parse_timestamp_with_minute(minute_mult, param_name, calendar, utc):
    minute = minute_mult
    minute_is_utc_ts = isinstance(minute, pd.Timestamp) and minute.tz is not None
    dt = m.parse_timestamp(minute, param_name, calendar, utc=utc)
    if not utc and not minute_is_utc_ts:
        assert dt == pd.Timestamp("2021-06-02 23:00")
    else:
        assert dt == pd.Timestamp("2021-06-02 23:00", tz=UTC)
    assert dt == dt.floor("min")


def test_parse_timestamp_with_second(second, sides, param_name):
    side = sides
    if side not in ["right", "left"]:
        error_msg = re.escape(
            "`timestamp` cannot have a non-zero second (or more accurate)"
            f" component for `side` '{side}'."
        )
        with pytest.raises(ValueError, match=error_msg):
            m.parse_timestamp(second, param_name, raise_oob=False, side=side)
    else:
        parsed = m.parse_timestamp(second, param_name, raise_oob=False, side=side)
        if side == "left":
            assert parsed == pd.Timestamp("2021-06-02 23:01", tz=UTC)
        else:
            assert parsed == pd.Timestamp("2021-06-02 23:02", tz=UTC)


def test_parse_timestamp_error_malformed(malformed, param_name, calendar):
    expected_error = TypeError if isinstance(malformed, tuple) else ValueError
    error_msg = re.escape(
        f"Parameter `{param_name}` receieved as '{malformed}' although a Date or"
        f" Minute must be passed as a pd.Timestamp or a valid single-argument"
        f" input to pd.Timestamp."
    )
    with pytest.raises(expected_error, match=error_msg):
        m.parse_timestamp(malformed, param_name, calendar)


def test_parse_timestamp_error_oob(
    calendar, param_name, minute_too_early, minute_too_late
):
    # by default raises error if oob
    error_msg = re.escape(
        f"Parameter `{param_name}` receieved as '{minute_too_early}' although"
        f" cannot be earlier than the first trading minute of calendar"
    )
    with pytest.raises(errors.MinuteOutOfBounds, match=error_msg):
        m.parse_timestamp(minute_too_early, param_name, calendar)

    # verify parses if oob and raise_oob False
    rtrn = m.parse_timestamp(minute_too_early, param_name, raise_oob=False, side="left")
    assert rtrn == minute_too_early

    # by default raises error if oob
    error_msg = re.escape(
        f"Parameter `{param_name}` receieved as '{minute_too_late}' although"
        f" cannot be later than the last trading minute of calendar"
    )
    with pytest.raises(errors.MinuteOutOfBounds, match=error_msg):
        m.parse_timestamp(minute_too_late, param_name, calendar)

    # verify parses if oob and raise_oob False
    rtrn = m.parse_timestamp(minute_too_late, param_name, raise_oob=False, side="left")
    assert rtrn == minute_too_late


def test_parse_date_or_minute_for_minute(
    calendar, param_name, minute, minute_mult, date
):
    """Tests `parse_date_or_minute` for input that represents a Minute."""

    def f(ts: pd.Timestamp) -> tuple[pd.Timestamp, bool]:
        return m.parse_date_or_minute(ts, param_name, calendar)

    assert f(minute_mult) == (pd.Timestamp(minute, tz=UTC), True)
    # verify that midnight with tz as UTC intepreted as minute, not date.
    assert f(pd.Timestamp(date, tz=UTC)) == (pd.Timestamp(date, tz=UTC), True)


def test_parse_date_or_minute_for_date(calendar, param_name, date, date_mult):
    """Tests `parse_date_or_minute` for input that represents a Minute."""
    f = m.parse_date_or_minute
    assert f(date_mult, param_name, calendar) == (pd.Timestamp(date), False)


def test_parse_date_or_minute_oob(
    calendar,
    param_name,
    date_too_early,
    date_too_late,
    minute_too_early,
    minute_too_late,
):
    """Tests `parse_date_or_minute` for out-of-bounds input.

    Tests as if an extension of parse_timestamp, i.e. only tests added
    functionality.
    """

    def f(ts: pd.Timestamp) -> tuple[pd.Timestamp, bool]:
        return m.parse_date_or_minute(ts, param_name, calendar)

    # Verify raises errors for out-of-bounds ts
    first_min = calendar.first_minute
    last_min = calendar.last_minute
    first_date = calendar.first_session
    last_date = calendar.last_session
    # verify returns at calendar's minute bounds
    assert f(first_min) == (first_min, True)
    assert f(last_min) == (last_min, True)
    # verify raises error other side of calendar's minute bounds
    with pytest.raises(errors.MinuteOutOfBounds):
        f(minute_too_early)
    with pytest.raises(errors.MinuteOutOfBounds):
        f(minute_too_late)
    # verify returns at calendar's session bounds
    assert f(first_date) == (first_date, False)
    assert f(last_date) == (last_date, False)
    # verify raises error other side of calendar's session bounds
    with pytest.raises(errors.DateOutOfBounds):
        f(date_too_early)
    with pytest.raises(errors.DateOutOfBounds):
        f(date_too_late)


def test_parse_date(date_mult, param_name):
    date = date_mult
    dt = m.parse_date(date, param_name, raise_oob=False)
    assert dt == pd.Timestamp("2021-06-05")


def test_parse_date_errors(calendar, param_name, date_too_early, date_too_late):
    dt = pd.Timestamp("2021-06-02", tz=ZoneInfo("America/Chicago"))
    with pytest.raises(ValueError, match="a Date must be timezone naive"):
        m.parse_date(dt, param_name, raise_oob=False)

    dt = pd.Timestamp("2021-06-02 13:33")
    with pytest.raises(ValueError, match="a Date must have a time component of 00:00."):
        m.parse_date(dt, param_name, raise_oob=False)

    # by default raises error if oob
    error_msg = (
        f"Parameter `{param_name}` receieved as '{date_too_early}' although"
        f" cannot be earlier than the first session of calendar"
    )
    with pytest.raises(errors.DateOutOfBounds, match=re.escape(error_msg)):
        m.parse_date(date_too_early, param_name, calendar)

    # verify parses if oob and raise_oob False
    assert m.parse_date(date_too_early, param_name, raise_oob=False) == date_too_early

    # by default parses oob
    error_msg = (
        f"Parameter `{param_name}` receieved as '{date_too_late}' although"
        f" cannot be later than the last session of calendar"
    )
    with pytest.raises(errors.DateOutOfBounds, match=re.escape(error_msg)):
        m.parse_date(date_too_late, param_name, calendar)

    # verify parses if oob and raise_oob False
    assert m.parse_date(date_too_late, param_name, raise_oob=False) == date_too_late


def test_parse_session(
    calendar, session, date, date_too_early, date_too_late, param_name
):
    ts = m.parse_session(calendar, session, param_name)
    assert ts == pd.Timestamp(session)

    with pytest.raises(errors.NotSessionError, match="not a session of calendar"):
        m.parse_session(calendar, date, param_name)

    with pytest.raises(
        errors.NotSessionError, match="is earlier than the first session of calendar"
    ):
        m.parse_session(calendar, date_too_early, param_name)

    with pytest.raises(
        errors.NotSessionError, match="is later than the last session of calendar"
    ):
        m.parse_session(calendar, date_too_late, param_name)


def test_parse_trading_minute(
    calendar, trading_minute, minute, minute_too_early, minute_too_late, param_name
):
    ts = m.parse_trading_minute(calendar, trading_minute, param_name)
    assert ts == pd.Timestamp(trading_minute, tz=UTC)

    with pytest.raises(
        errors.NotTradingMinuteError, match="not a trading minute of calendar"
    ):
        m.parse_trading_minute(calendar, minute, param_name)

    with pytest.raises(
        errors.NotTradingMinuteError,
        match="is earlier than the first trading minute of calendar",
    ):
        m.parse_trading_minute(calendar, minute_too_early, param_name)

    with pytest.raises(
        errors.NotTradingMinuteError,
        match="is later than the last trading minute of calendar",
    ):
        m.parse_trading_minute(calendar, minute_too_late, param_name)


def st_align() -> st.SearchStrategy[pd.Timedelta]:
    """SearchStrategy for a valid alignment."""
    sample_pos = [pd.Timedelta(i, "min") for i in range(1, 31) if not 60 % i]
    sample_neg = [-td for td in sample_pos]
    return st.sampled_from(sample_pos + sample_neg)


class TestTradingIndex:
    """Tests for _TradingIndex.

    Subjects selected calendars (each of a unique behaviour) to fuzz tests
    verifying expected behaviour / no unexpected errors. These tests cover
    all date ranges, periods (from 1 minute to 1 day) and options.

    Also includes:
        - concrete tests to verify overlap handling.
        - concrete test to verify passing start and/or end as a time.
        - parsing tests for ExchangeCalendar.trading_index.

    NOTE: `_TradingIndex` is also tested via
    `ExchangeCalendarTestBase.test_trading_index` which tests a multitude
    of concrete cases (options as default values).
    """

    CALENDAR_NAMES = ["XLON", "XHKG", "XTAE", "CMES", "24/7"]
    """Selection of calendars with a particular behaviour:
    "XLON" - calendars without breaks.
    "XHKG" - calendars with breaks.
    "XTAE" - opens at 9:59am, useful for testing `align`
    "CMES" - 24 hour calendar, not 7 days a week.
    "24/7" - 24 hour calendar.
    """

    # Fixtures

    @pytest.fixture(scope="class")
    def answers(self) -> abc.Iterator[dict[str, Answers]]:
        """Dict of answers for tested calendars, key as name, value as Answers."""
        d = {}
        for name in self.CALENDAR_NAMES:
            d[name] = Answers(name, side="left")
        return d

    @pytest.fixture(scope="class")
    def calendars(self, answers) -> abc.Iterator[dict[str, ExchangeCalendar]]:
        """Dict of tested calendars, key as name, value as calendar."""
        d = {}
        for name, ans in answers.items():
            cls = calendar_utils._default_calendar_factories[name]
            d[name] = cls(start=ans.first_session, end=ans.last_session)
        return d

    @pytest.fixture(scope="class", params=CALENDAR_NAMES)
    def calendars_with_answers(
        self, request, calendars, answers
    ) -> abc.Iterator[tuple[ExchangeCalendar, Answers]]:
        """Parameterized fixture."""
        yield (calendars[request.param], answers[request.param])

    # Helper strategies

    @staticmethod
    @st.composite
    def _st_times_different(draw, ans) -> tuple[pd.Timestamp, pd.Timestamp]:
        """SearchStrategy for two consecutive sessions with different times."""
        session = draw(st.sampled_from(ans.sessions_next_time_different.to_list()))
        next_session = ans.get_next_session(session)
        return (session, next_session)

    @staticmethod
    @st.composite
    def _st_start_end(draw, ans) -> tuple[pd.Timestamp, pd.Timestamp]:
        """SearchStrategy for start and end dates in calendar range and
        a calendar specific maximum distance."""
        first = ans.first_session
        last = ans.last_session

        one_day = pd.Timedelta(1, "D")
        # reasonable to quicken test by limiting 24/7 as rules for 24/7 are unchanging.
        distance = (
            pd.DateOffset(weeks=2) if ans.name == "24/7" else pd.DateOffset(years=1)
        )

        end = draw(st.datetimes(min(first + distance, last), last))
        end = pd.Timestamp(end).floor("D")
        start = draw(st.datetimes(max(end - distance, first), end - one_day))
        start = pd.Timestamp(start).floor("D")
        start, end = start, end
        assume(not ans.answers[start:end].empty)
        return start, end

    def st_start_end(
        self, ans: Answers
    ) -> st.SearchStrategy[tuple[pd.Timestamp, pd.Timestamp]]:
        """SearchStrategy for trading index start and end dates."""
        st_startend = self._st_start_end(ans)
        if not ans.sessions_next_time_different.empty:
            st_times_differ = self._st_times_different(ans)
            st_startend = st.one_of(st_startend, st_times_differ)
        return st_startend

    @staticmethod
    @st.composite
    def st_periods(
        draw,
        minimum: pd.Timedelta = pd.Timedelta(1, "min"),
        maximum: pd.Timedelta = pd.Timedelta(1, "D") - pd.Timedelta(1, "min"),
    ) -> pd.Timedelta:
        """SearchStrategy for a period between a `minimum` and `maximum`."""
        period = draw(st.integers(minimum.seconds // 60, maximum.seconds // 60))
        return pd.Timedelta(period, "min")

    # Helper methods

    @staticmethod
    def could_overlap(
        ans: Answers,
        slc: slice,
        has_break: bool,
        align: pd.Timedelta,
        align_pm: pd.Timedelta,
    ) -> bool:
        """Query if there's at least one period at which intervals overlap.

        Can right side of last interval of any session/subsession of a
        slice of Answers fall later than the left side of the first
        interval of the next session/subsession?
        """
        can_overlap = False
        if has_break:
            duration = ans.break_starts[slc] - ans.opens[slc].dt.ceil(align)
            gap = ans.break_ends[slc].dt.ceil(align_pm) - ans.break_starts[slc]
            can_overlap = (gap < duration).any()
        if not can_overlap:
            duration = ans.closes[slc] - ans.opens[slc].dt.ceil(align)
            gap = ans.opens.shift(-1)[slc].dt.ceil(align) - ans.closes[slc]
            can_overlap = (gap < duration).any()
        return can_overlap

    @staticmethod
    def evaluate_overrun(
        starts: pd.Series,
        ends: pd.Series,
        period: pd.Timedelta,
    ) -> pd.Series:
        """Evaluate extent that right side of last interval exceeds end.

        For session/subsessions with `starts` and `ends` evaluate the
        distance beyond the `ends` that the right of the last interval will
        fall (where interval length is `period`).
        """
        duration = ends - starts
        shortfall = duration % period
        on_end_mask = shortfall == pd.Timedelta(0)
        overrun = period - shortfall
        overrun[on_end_mask] = pd.Timedelta(0)
        return overrun

    @staticmethod
    def sessions_bounds(
        ans: Answers,
        slc: slice,
        period: pd.Timedelta,
        closed: str | None,
        force_break_close: bool,
        force_close: bool,
        align: pd.Timedelta,
        align_pm: pd.Timedelta,
        curtail: bool = False,
    ) -> tuple[pd.Series, pd.Series]:
        """First and last trading indice of each session/subsession.

        `closed` should be passed as None if evaluating bounds for an
        intervals index.
        """
        closed_left = closed in [None, "left", "both"]
        closed_right = closed in [None, "right", "both"]

        opens = ans.opens[slc]
        closes = ans.closes[slc]
        has_break = ans.break_starts[slc].notna().any()

        def bounds(start: pd.Series, end: pd.Series, force: bool, align: pd.Timedelta):
            """Evaluate bounds of trading index by session/subsession.

            Parameters
            ----------
            start
                Index: pd.DatetimeIndex
                    session. Must be the same as `end.index`.
                Value: pd.DatetimeIndex
                    Start time of session or a subsession of session (where
                    session is index value).

            end
                As for `start` albeit indicating end times.
            """
            start = start.dt.ceil(align) if align else start
            lower_bounds = start if closed_left else start + period
            if force and closed_right:
                if (lower_bounds > end).any():
                    # period longer than session/subsession duration
                    lower_bounds[lower_bounds > end] = end
                return lower_bounds, end

            duration = end - start
            func = np.ceil if closed_right else np.floor
            num_periods = func(duration / period)
            if not closed_right and closed is not None:
                num_periods[duration % period == pd.Timedelta(0)] -= 1

            upper_bounds = start + (num_periods * period)

            if closed == "neither" and (num_periods == 0).any:  # edge case
                # lose bounds where session/subsession has no indice.
                upper_bounds = upper_bounds[num_periods != 0]
                lower_bounds = lower_bounds[upper_bounds.index]

            return lower_bounds, upper_bounds

        if has_break:
            break_starts = ans.break_starts[slc]
            break_ends = ans.break_ends[slc]
            mask = break_starts.notna()  # which sessions have a break

            # am sessions bounds
            am_lower, am_upper = bounds(
                opens[mask], break_starts[mask], force_break_close, align
            )

            # pm sessions bounds
            pm_lower, pm_upper = bounds(
                break_ends[mask], closes[mask], force_close, align_pm
            )

            # sessions without breaks
            if (~mask).any():
                day_lower, day_upper = bounds(
                    opens[~mask], closes[~mask], force_close, align
                )
            else:
                day_upper = day_lower = pd.Series([], dtype="datetime64[ns, UTC]")

            lower_bounds = pd.concat(
                [srs for srs in (am_lower, pm_lower, day_lower) if not srs.empty]
            )
            upper_bounds = pd.concat(
                [srs for srs in (am_upper, pm_upper, day_upper) if not srs.empty]
            )

        else:
            lower_bounds, upper_bounds = bounds(opens, closes, force_close, align)

        if curtail and not (force_close and force_break_close):
            indices = lower_bounds.argsort()
            lower_bounds = lower_bounds.sort_values()
            upper_bounds = upper_bounds.iloc[indices]
            curtail_mask = upper_bounds > lower_bounds.shift(-1)
            if curtail_mask.any():
                upper_bounds[curtail_mask] = lower_bounds.shift(-1)[curtail_mask]

        return lower_bounds, upper_bounds

    # Fuzz tests for unexpected errors and return behaviour.

    @given(
        data=st.data(),
        force_close=st.booleans(),
        force_break_close=st.booleans(),
        align=st_align(),
        align_pm=st_align(),
    )
    @settings(deadline=None, suppress_health_check=[HealthCheck.differing_executors])
    def test_indices_fuzz(
        self,
        data,
        calendars_with_answers,
        force_close: bool,
        force_break_close: bool,
        align,
        align_pm,
        one_min,
    ):
        """Fuzz for unexpected errors and options behaviour.

        Expected errors tested for separately.

        'period' limited to avoid IndicesOverlapError.

        'start' and 'end' set to:
            be within calendar bounds.
            dates covering at least one session.
        """
        cal, ans = calendars_with_answers
        start, end = data.draw(self.st_start_end(ans))
        slc = ans.sessions.slice_indexer(start, end)
        has_break = ans.break_starts[slc].notna().any()

        closed_options = ["left", "neither", "right", "both"]
        if ans.sessions[slc].isin(ans.sessions_without_gap_after).any():
            closed_options = closed_options[:-1]  # lose "both" option
        closed = data.draw(st.sampled_from(closed_options))
        closed_right = closed in ["right", "both"]

        max_period = pd.Timedelta(1, "D") - one_min

        params_allow_overlap = closed_right and not (force_break_close and force_close)
        if params_allow_overlap:
            can_overlap = self.could_overlap(ans, slc, has_break, align, align_pm)
        else:
            can_overlap = False

        if has_break and can_overlap:
            # filter out periods that will definitely overlap.
            max_period = (
                ans.break_ends[slc].dt.ceil(align_pm) - ans.opens[slc].dt.ceil(align)
            ).min()

        # guard against "neither" returning empty. Tested for under seprate test.
        if closed == "neither":
            if has_break:
                am_length = (
                    ans.break_starts[slc] - ans.opens[slc].dt.ceil(align)
                ).min() - one_min
                pm_length = (
                    ans.closes[slc] - ans.break_ends[slc].dt.ceil(align_pm)
                ).min() - one_min
                max_period = min(max_period, am_length, pm_length)
            else:
                min_length = (
                    ans.closes[slc] - ans.opens[slc].dt.ceil(align)
                ).min() - one_min
                max_period = min(max_period, min_length)

        period = data.draw(self.st_periods(maximum=max_period))

        if can_overlap:
            # assume no overlaps (i.e. reject test parameters if would overlap).
            op = operator.ge if closed == "both" else operator.gt
            if has_break:
                mask = ans.break_starts[slc].notna()
                overrun = self.evaluate_overrun(
                    ans.opens[slc][mask].dt.ceil(align),
                    ans.break_starts[slc][mask],
                    period,
                )
                break_ends_aligned = ans.break_ends[slc].dt.ceil(align_pm)
                break_duration = (break_ends_aligned - ans.break_starts[slc]).dropna()
                assume(not op(overrun, break_duration).any())
            overrun = self.evaluate_overrun(
                ans.opens[slc].dt.ceil(align), ans.closes[slc], period
            )
            sessions_gap = ans.opens[slc].shift(-1).dt.ceil(align) - ans.closes[slc]
            assume(not op(overrun, sessions_gap).any())

        ti = m._TradingIndex(
            cal,
            start,
            end,
            period,
            closed,
            force_close,
            force_break_close,
            curtail_overlaps=False,
            ignore_breaks=False,
            align=align,
            align_pm=align_pm,
        )
        index = ti.trading_index()

        # Assertions

        assert isinstance(index, pd.DatetimeIndex)
        assert not index.empty

        lower_bounds, upper_bounds = self.sessions_bounds(
            ans, slc, period, closed, force_break_close, force_close, align, align_pm
        )

        assert lower_bounds.isin(index).all()
        assert upper_bounds.isin(index).all()

        # verify that all indices are within bounds of a session or subsession.
        bv = pd.Series(False, index)
        for lower_bound, upper_bound in zip(lower_bounds, upper_bounds):
            bv = bv | ((index >= lower_bound) & (index <= upper_bound))
        assert bv.all()

    @given(
        data=st.data(),
        force_break_close=st.booleans(),
        curtail=st.booleans(),
        align=st_align(),
        align_pm=st_align(),
    )
    @settings(deadline=None, suppress_health_check=[HealthCheck.differing_executors])
    def test_intervals_fuzz(
        self,
        data,
        calendars_with_answers,
        force_break_close: bool,
        curtail: bool,
        align: pd.Timedelta,
        align_pm: pd.Timedelta,
        one_min,
    ):
        """Fuzz for unexpected errors and options behaviour.

        Expected errors tested for separately.

        `period` limited to avoid IntervalsOverlapError.

        'start' and 'end' set to:
            be within calendar bounds.
            dates covering at least one session.
        """
        cal, ans = calendars_with_answers
        start, end = data.draw(self.st_start_end(ans))
        slc = ans.sessions.slice_indexer(start, end)
        has_break = ans.break_starts[slc].notna().any()

        force_close = data.draw(st.booleans())
        closed = data.draw(st.sampled_from(["left", "right"]))
        max_period = pd.Timedelta(1, "D") - one_min

        params_allow_overlap = not curtail and not (force_break_close and force_close)
        if params_allow_overlap:
            can_overlap = self.could_overlap(ans, slc, has_break, align, align_pm)
        else:
            can_overlap = False

        if has_break and can_overlap:
            # filter out periods that will definitely overlap.
            max_period = (ans.break_ends[slc] - ans.opens[slc]).min()

        period = data.draw(self.st_periods(maximum=max_period))

        if can_overlap:
            # assume no overlaps
            if has_break:
                mask = ans.break_starts[slc].notna()
                overrun = self.evaluate_overrun(
                    ans.opens[slc][mask].dt.ceil(align),
                    ans.break_starts[slc][mask],
                    period,
                )
                break_ends_aligned = ans.break_ends[slc].dt.ceil(align_pm)
                break_duration = (break_ends_aligned - ans.break_starts[slc]).dropna()
                assume(not (overrun > break_duration).any())
            overrun = self.evaluate_overrun(
                ans.opens[slc].dt.ceil(align), ans.closes[slc], period
            )
            sessions_gap = ans.opens[slc].shift(-1).dt.ceil(align) - ans.closes[slc]
            assume(not (overrun > sessions_gap).any())

        ti = m._TradingIndex(
            cal,
            start,
            end,
            period,
            closed,
            force_close,
            force_break_close,
            curtail,
            ignore_breaks=False,
            align=align,
            align_pm=align_pm,
        )
        index = ti.trading_index_intervals()

        # assertions

        assert isinstance(index, pd.IntervalIndex)
        assert not index.empty

        lower_bounds, upper_bounds = self.sessions_bounds(
            ans,
            slc,
            period,
            None,
            force_break_close,
            force_close,
            align,
            align_pm,
            curtail,
        )

        assert lower_bounds.isin(index.left).all()
        assert upper_bounds.isin(index.right).all()

        # verify that all intervals are within bounds of a session or subsession
        bv = pd.Series(False, index)
        for lower_bound, upper_bound in zip(lower_bounds, upper_bounds):
            bv = bv | ((index.left >= lower_bound) & (index.right <= upper_bound))

    @given(data=st.data(), calendar_name=st.sampled_from(["XLON", "XHKG"]))
    @settings(deadline=None)
    def test_for_empty_with_neither_fuzz(
        self, data, calendars, answers, calendar_name, one_min
    ):
        """Fuzz for specific condition that returns empty DatetimeIndex.

        Fuzz for expected empty DatetimeIndex when closed "neither" and
        period is longer than any session/subsession.
        """
        cal, ans = calendars[calendar_name], answers[calendar_name]
        start, end = data.draw(self.st_start_end(ans))
        slc = ans.sessions.slice_indexer(start, end)
        has_break = ans.break_starts[slc].notna().any()

        if has_break:
            max_am_length = (ans.break_starts[slc] - ans.opens[slc]).max()
            max_pm_length = (ans.closes[slc] - ans.break_ends[slc]).max()
            min_period = max(max_am_length, max_pm_length)
        else:
            min_period = (ans.closes[slc] - ans.opens[slc]).max()

        period = data.draw(self.st_periods(minimum=min_period))

        closed = "neither"
        forces = [False, False]

        ti = m._TradingIndex(
            cal, start, end, period, closed, *forces, False, False, one_min, one_min
        )
        index = ti.trading_index()
        assert index.empty

    @given(
        data=st.data(),
        intervals=st.booleans(),
        force_close=st.booleans(),
        force_break_close=st.booleans(),
        curtail_overlaps=st.booleans(),
    )
    @settings(deadline=None, suppress_health_check=[HealthCheck.differing_executors])
    def test_daily_fuzz(
        self,
        data,
        calendars_with_answers,
        intervals: bool,
        force_close: bool,
        force_break_close: bool,
        curtail_overlaps: bool,
    ):
        """Fuzz for unexpected errors and return behaviour."""
        cal, ans = calendars_with_answers

        if intervals:
            closed_options = ["left", "right"]
        else:
            closed_options = ["left", "right", "neither", "both"]
        closed = data.draw(st.sampled_from(closed_options))

        start, end = data.draw(self.st_start_end(ans))
        period = pd.Timedelta(1, "D")
        forces_and_curtails = [force_close, force_break_close, None, curtail_overlaps]

        index = cal.trading_index(
            start, end, period, intervals, closed, *forces_and_curtails, parse=False
        )
        assert isinstance(index, pd.DatetimeIndex)
        assert not index.empty
        pd.testing.assert_index_equal(index.normalize(), index)

    # Tests for expected errors.

    @pytest.mark.parametrize("name", ["XHKG", "24/7", "CMES"])
    @given(data=st.data(), closed=st.sampled_from(["right", "both"]))
    @settings(deadline=None, suppress_health_check=[HealthCheck.differing_executors])
    def test_overlap_error_fuzz(self, data, name, calendars, answers, closed, one_min):
        """Fuzz for expected IndicesOverlapError.

        NB. Test should exclude calendars, such as "XLON", for which
        indices cannot overlap. These are calendars where a
        session/subsession duration is less than the subsequent gap
        between that session/subsession and the next. Passing any slice of
        the answers for such a calendar to `could_overlap` would return
        False. That such calendars cannot have overlapping indices is
        verified by `test_indices_fuzz` and `test_intervals_fuzz` which
        place no restraints on the period that these calendars can be
        tested against (at least between 0 minutes and 1 day exclusive).
        """
        cal, ans = calendars[name], answers[name]
        start, end = data.draw(self.st_start_end(ans))
        slc = ans.sessions.slice_indexer(start, end)
        has_break = ans.break_starts[slc].notna().any()

        # filter out periods that will definitely not cause an overlap.
        if has_break:
            min_period = (ans.break_ends[slc] - ans.break_starts[slc]).min()
        else:
            min_period = (ans.opens.shift(-1)[slc] - ans.closes[slc]).min()

        period = data.draw(self.st_periods(minimum=max(one_min, min_period)))

        # assume overlaps (i.e. reject test parameters if does not overlap)
        op = operator.ge if closed == "both" else operator.gt
        if has_break:
            mask = ans.break_starts[slc].notna()
            overrun = self.evaluate_overrun(
                ans.opens[slc][mask], ans.break_starts[slc][mask], period
            )
            break_duration = (ans.break_ends[slc] - ans.break_starts[slc]).dropna()
            assume(op(overrun, break_duration).any())
        else:
            overrun = self.evaluate_overrun(ans.opens[slc], ans.closes[slc], period)
            sessions_gap = ans.opens[slc].shift(-1) - ans.closes[slc]
            assume(op(overrun, sessions_gap).any())

        ti = m._TradingIndex(
            cal,
            start,
            end,
            period,
            closed,
            force_close=False,
            force_break_close=False,
            curtail_overlaps=False,
            ignore_breaks=False,
            align=one_min,
            align_pm=one_min,
        )
        with pytest.raises(errors.IndicesOverlapError):
            ti.trading_index()

        if closed == "right":
            with pytest.raises(errors.IntervalsOverlapError):
                ti.trading_index_intervals()

    @pytest.fixture(params=[True, False])
    def curtail_all(self, request) -> abc.Iterator[bool]:
        """Parameterized fixture of all values for 'curtail_overlaps'."""
        yield request.param

    @pytest.fixture(scope="class")
    def cal_start_end(
        self, calendars
    ) -> abc.Iterator[tuple[ExchangeCalendar], pd.Timestamp, pd.Timestamp]:
        """(calendar, start, end) parameters for specific tests."""
        yield (
            calendars["XHKG"],
            pd.Timestamp("2018-01-01"),
            pd.Timestamp("2018-12-31"),
        )

    @pytest.fixture(params=itertools.product(("105min", "106min"), ("right", "both")))
    def ti_for_overlap(
        self, request, cal_start_end, curtail_all, one_min
    ) -> abc.Iterator[m._TradingIndex]:
        """_TradingIndex fixture against which to test for overlaps.

        '105T' is the edge case where last right indice of am subsession would
        coincide with first left indice of pm subsession.
        '106T' is one minute to the right of this.
        """
        cal, start, end = cal_start_end
        period, closed = request.param
        period = pd.Timedelta(period)
        yield m._TradingIndex(
            cal,
            start,
            end,
            period,
            closed,
            force_close=False,
            force_break_close=False,
            curtail_overlaps=curtail_all,
            ignore_breaks=False,
            align=one_min,
            align_pm=one_min,
        )

    def test_overlaps(self, ti_for_overlap, answers):
        """Test 'curtail_overlaps' and for overlaps against concrete parameters."""
        ti = ti_for_overlap
        period = pd.Timedelta(ti.interval_nanos)
        period_106 = period == pd.Timedelta("106min")

        if period_106 or ti.closed == "both":
            with pytest.raises(errors.IndicesOverlapError):
                ti.trading_index()

        if ti.closed == "both":
            return

        if not ti.curtail_overlaps and period_106:
            # won't raise on "105min" as right side of last interval of am
            # session won't clash on coinciding with left side of first
            # interval of pm session as one of these sides will always be
            # open (in this case the left side). NB can't close on both
            # sides.
            with pytest.raises(errors.IntervalsOverlapError):
                ti.trading_index_intervals()
        else:
            index = ti.trading_index_intervals()
            assert index.is_non_overlapping_monotonic
            assert index.right.isin(answers["XHKG"].break_ends).any()

    @pytest.fixture(params=("right", "both"))
    def ti_for_overlap_error_negative_case(
        self, request, cal_start_end, curtail_all, one_min
    ) -> abc.Iterator[m._TradingIndex]:
        """_TradingIndex fixture against which to test for no overlaps.

        104T is the edge case, one minute short of coinciding with pm subsession open.
        """
        cal, start, end = cal_start_end
        period, closed = pd.Timedelta("104min"), request.param
        yield m._TradingIndex(
            cal,
            start,
            end,
            period,
            closed,
            force_close=False,
            force_break_close=False,
            curtail_overlaps=curtail_all,
            ignore_breaks=True,
            align=one_min,
            align_pm=one_min,
        )

    def test_overlaps_2(self, ti_for_overlap_error_negative_case):
        """Test for no overlaps against concrete edge case."""
        ti = ti_for_overlap_error_negative_case
        index = ti.trading_index()
        assert isinstance(index, pd.DatetimeIndex)
        if ti.closed == "right":
            index = ti.trading_index_intervals()
            assert isinstance(index, pd.IntervalIndex)
            assert index.is_non_overlapping_monotonic

    # Tests for other options

    def test_force(self, cal_start_end):
        """Verify `force` option overrides `force_close` and `force_break_close`."""
        cal, start, end = cal_start_end
        kwargs = dict(start=start, end=end, period="1h", intervals=True)

        expected_true = cal.trading_index(
            **kwargs, force_close=True, force_break_close=True
        )
        expected_false = cal.trading_index(
            **kwargs, force_close=False, force_break_close=False
        )

        force_combos = itertools.product([True, False], repeat=2)
        for force_close, force_break_close in force_combos:
            rtrn_true = cal.trading_index(
                **kwargs,
                force_close=force_close,
                force_break_close=force_break_close,
                force=True,
            )

            assert_index_equal(rtrn_true, expected_true)

            rtrn_false = cal.trading_index(
                **kwargs,
                force_close=force_close,
                force_break_close=force_break_close,
                force=False,
            )

            assert_index_equal(rtrn_false, expected_false)

    def test_ignore_breaks(self, cal_start_end):
        """Verify effect of ignore_breaks option."""
        cal, start, end = cal_start_end
        assert cal.sessions_has_break(start, end)
        kwargs = dict(start=start, end=end, period="1h", intervals=True)

        # verify a difference
        index_false = cal.trading_index(**kwargs, ignore_breaks=False)
        index_true = cal.trading_index(**kwargs, ignore_breaks=True)
        with pytest.raises(AssertionError):
            assert_index_equal(index_false, index_true)

        # create an equivalent calendar without breaks
        delta = pd.Timedelta(10, "D")
        start_ = start - delta
        end_ = end + delta

        cal_amended = calendar_utils._default_calendar_factories[cal.name](start_, end_)
        cal_amended.break_starts_nanos[:] = pd.NaT.value
        cal_amended.break_ends_nanos[:] = pd.NaT.value
        cal_amended.schedule.loc[:, "break_start"] = pd.NaT
        cal_amended.schedule.loc[:, "break_end"] = pd.NaT

        # verify amended calendar returns as original with breaks ignored
        rtrn = cal_amended.trading_index(**kwargs, ignore_breaks=False)
        assert_index_equal(rtrn, index_true)

    @pytest.fixture(scope="class")
    def cal_with_ans_align(self) -> abc.Iterator[tuple[ExchangeCalendar, Answers]]:
        """Calendar with open and break_end times to test align options."""
        cal_name = "TEST"

        # NOTE Changing the timings of this test calendar class also requires
        # changing the corresponding test.csv answers file in the resources dir.
        class TESTCal(XTAEExchangeCalendar):
            name = cal_name
            break_start_times = ((None, time(15, 58)),)
            break_end_times = ((None, time(17, 28)),)
            close_times = ((None, time(19, 15)),)

        ans = Answers(cal_name, "left")
        cal = TESTCal(start=ans.first_session, end=ans.last_session)
        yield cal, ans

    @pytest.fixture(scope="class")
    def dates_align(
        self, cal_with_ans_align
    ) -> abc.Iterator[tuple[pd.Timestamp, pd.Timestamp]]:
        """Sessions over which to test effect of align parameters.

        Two contiguous sessions with times as asserted.
        """
        _, ans = cal_with_ans_align
        from_, to = pd.Timestamp("2020-12-10"), pd.Timestamp("2020-12-13")
        # assert assumed open / close times
        assert ans.opens[from_].time() == time(7, 59)
        assert ans.break_starts[from_].time() == time(13, 58)
        assert ans.break_ends[from_].time() == time(15, 28)
        assert ans.closes[from_].time() == time(17, 15)
        assert ans.opens[to].time() == time(7, 59)
        assert ans.break_starts[to] is ans.break_ends[to] is pd.NaT
        assert ans.closes[to].time() == time(13, 40)
        yield from_, to

    @given(
        data=st.data(),
        force_close=st.booleans(),
        force_break_close=st.booleans(),
        closed=st.sampled_from(["left", "right", "both", "neither"]),
        ignore_breaks=st.booleans(),
    )
    @settings(deadline=None)
    def test_align(
        self,
        data,
        cal_with_ans_align,
        dates_align,
        closed,
        force_close,
        force_break_close,
        ignore_breaks,
        one_min,
    ):
        """Test `align` option.

        Additional concrete test to verify effect of `align` and `align_pm' on
        'alignable' calendar. (Effect of these options is principally tested
        by `test_indices_fuzz` and `test_intervals_fuzz`.)
        """
        cal, ans = cal_with_ans_align
        from_, to = dates_align
        aligned_start_times = {
            "-1min": (time(7, 59), time(15, 28)),
            "-2min": (time(7, 58), time(15, 28)),
            "-3min": (time(7, 57), time(15, 27)),
            "-5min": (time(7, 55), time(15, 25)),
            "-15min": (time(7, 45), time(15, 15)),
            "-20min": (time(7, 40), time(15, 20)),
            "-30min": (time(7, 30), time(15)),
            "-60min": (time(7), time(15)),
            "1min": (time(7, 59), time(15, 28)),
            "2min": (time(8), time(15, 28)),
            "3min": (time(8), time(15, 30)),
            "5min": (time(8), time(15, 30)),
            "15min": (time(8), time(15, 30)),
            "20min": (time(8), time(15, 40)),
            "30min": (time(8), time(15, 30)),
            "60min": (time(8), time(16)),
        }
        alignments = list(aligned_start_times.keys())
        align = data.draw(st.sampled_from(alignments))
        align_pm = data.draw(st.one_of([st.sampled_from(alignments), st.booleans()]))
        period = data.draw(self.st_periods(maximum=pd.Timedelta(1, "h")))

        open_pm = ans.break_ends[from_]
        closes = (ans.break_starts[from_], ans.closes[from_], ans.closes[to])
        closes_ignore_breaks = (ans.closes[from_], ans.closes[to])

        closed_left = closed in ["left", "both"]
        closed_right = closed in ["right", "both"]
        tz = UTC

        def create_expected(
            starts: list[pd.Timestamp],
            ends: list[pd.Timestamp],
            period: pd.Timedelta,
            forces: list[bool],
        ) -> pd.DatetimeIndex:
            index_ = pd.DatetimeIndex([], tz=tz)
            for start, end, force in zip(starts, ends, forces):
                index = pd.date_range(start, end, freq=period, tz=tz)
                if not closed_left:
                    index = index[1:]
                if closed_right and end != index[-1]:
                    index = index.insert(len(index), index[-1] + period)
                if not closed_right and end == index[-1]:
                    index = index[:-1]
                if force and index[-1] > end:
                    index = index[:-1].insert(len(index) - 1, end)
                index_ = index_.union(index)
            return index_

        def create_expected_intervals(
            starts: list[pd.Timestamp],
            ends: list[pd.Timestamp],
            period: pd.Timedelta,
            forces: list[bool],
        ) -> pd.DatetimeIndex:
            left_ = pd.DatetimeIndex([], tz=tz)
            right_ = pd.DatetimeIndex([], tz=tz)
            for start, end, force in zip(starts, ends, forces):
                left = pd.date_range(start, end - one_min, freq=period, tz=tz)
                right = left + period
                if force and right[-1] > end:
                    right = right[:-1].insert(len(right) - 1, end)
                left_ = left_.union(left)
                right_ = right_.union(right)
            return pd.IntervalIndex.from_arrays(left_, right_, closed)

        combine = datetime.datetime.combine

        def get_start(date: pd.Timestamp, tm: time):
            return pd.Timestamp(combine(date.date(), tm), tz=UTC)

        aligned_time_am, _ = aligned_start_times[align]
        starts = [get_start(from_, aligned_time_am)]
        if not ignore_breaks:
            if align_pm:
                alignment_pm = align if align_pm is True else align_pm
                _, aligned_time_pm = aligned_start_times[alignment_pm]
                start = get_start(from_, aligned_time_pm)
            else:
                start = open_pm
            starts.append(start)
        starts.append(get_start(to, aligned_time_am))
        ends = closes_ignore_breaks if ignore_breaks else closes
        forces = (
            [force_close, force_close]
            if ignore_breaks
            else [force_break_close, force_close, force_close]
        )

        args = (from_, to, period)
        kwargs = dict(
            closed=closed,
            align=align,
            align_pm=align_pm,
            ignore_breaks=ignore_breaks,
            force_close=force_close,
            force_break_close=force_break_close,
        )

        intervals = False
        expected = create_expected(starts, ends, period, forces)
        rtrn = cal.trading_index(*args, intervals=intervals, **kwargs)
        assert_index_equal(rtrn, expected)
        # test passing start and end as times as opposed to sessions...
        alt_args = (expected[0] + one_min, expected[-1] - one_min, period)
        rtrn = cal.trading_index(*alt_args, intervals=intervals, **kwargs)
        assert_index_equal(rtrn, expected[1:-1])

        intervals = True
        if closed not in ["left", "right"]:
            return
        expected = create_expected_intervals(starts, ends, period, forces)
        rtrn = cal.trading_index(*args, intervals=intervals, **kwargs)
        # test passing start and end as times as opposed to sessions...
        alt_args = (expected[0].left + one_min, expected[-1].right - one_min, period)
        rtrn = cal.trading_index(*alt_args, intervals=intervals, **kwargs)
        assert_index_equal(rtrn, expected[1:-1])

    def test_align_overlap(self, cal_with_ans_align, dates_align, one_min):
        """Test align options can cause overlap error.

        Tests concrete case raises overlap errors due to alignment
        regardless that period is shorter than the break.
        """
        cal, _ = cal_with_ans_align

        kwargs = dict(closed="right", align="-5min")
        align_pm = "-1h"

        intervals = True
        # assert returns at edge
        period = pd.Timedelta(85, "min")
        rtrn = cal.trading_index(
            *dates_align, period, intervals=intervals, align_pm=align_pm, **kwargs
        )
        limit = pd.Timestamp(
            datetime.datetime.combine(dates_align[0], time(15)), tz=UTC
        )
        assert limit in rtrn.right

        # assert raises beyond edge
        period += one_min
        with pytest.raises(errors.IntervalsOverlapError):
            cal.trading_index(
                *dates_align, period, intervals=intervals, align_pm=align_pm, **kwargs
            )

        # assert returns beyond edge if curtail
        rtrn = cal.trading_index(
            *dates_align,
            period,
            intervals=intervals,
            align_pm=align_pm,
            curtail_overlaps=True,
            **kwargs,
        )
        assert not rtrn.empty

        # assert returns beyond edge if no pm alignment
        rtrn = cal.trading_index(
            *dates_align, period, intervals=intervals, align_pm=False, **kwargs
        )
        assert not rtrn.empty

        intervals = False
        # assert returns at edge
        period = pd.Timedelta(85, "min")
        rtrn = cal.trading_index(
            *dates_align, period, intervals=intervals, align_pm=align_pm, **kwargs
        )
        assert not rtrn.empty

        # assert raises beyond edge
        period += one_min
        with pytest.raises(errors.IndicesOverlapError):
            cal.trading_index(
                *dates_align, period, intervals=intervals, align_pm=align_pm, **kwargs
            )

        # assert returns beyond edge if no pm alignment
        rtrn = cal.trading_index(
            *dates_align, period, intervals=intervals, align_pm=False, **kwargs
        )
        assert not rtrn.empty

    def test_start_end_times(self, one_min, calendars):
        """Test effect of passing start and/or end as a time.

        Tests passing start / end as combinations of dates and/or times.

        Tests by comparing return with subset of return for start and end
        as sessions.

        Tests return with `intervals` as True (IntervalIndex) and False
        (DatetimeIndex). With `intervals` as False tests for all `closed`
        options.
        """
        cal = calendars["XHKG"]

        # Define a start session and end session as sessions of standard length
        start_s = pd.Timestamp("2021-12-06")
        end_s = pd.Timestamp("2021-12-20")

        # assert of standard length
        standard_length = pd.Timedelta(hours=6, minutes=30)
        start_s_open, start_s_close = cal.session_open_close(start_s)
        assert start_s_close - start_s_open == standard_length
        end_s_open, end_s_close = cal.session_open_close(end_s)
        assert end_s_close - end_s_open == standard_length

        f = cal.trading_index

        # Note: when intervals = False the return will include the indice coinciding
        # with `start` and `end` even if the period that indice represents falls,
        # respectively, before `start` or after `end`

        def assertions(
            starts: list[
                tuple[
                    pd.Timestamp,
                    int | None,
                    int | None,
                    int | None,
                    int | None,
                    int | None,
                ]
            ],
            ends: list[
                tuple[
                    pd.Timestamp,
                    int | None,
                    int | None,
                    int | None,
                    int | None,
                    int | None,
                ]
            ],
            period: pd.Timedelta | str,
            force: bool,
            ignore_breaks: bool,
            curtail_overlaps: bool = False,
        ):
            """Assert returns slice of return for sessions.

            Parameters
            ----------
            starts: list of tuple (see method signature) of:
                [0] value to pass as start.
                Other items define the start of slice of subset when start is [0] and:
                [1] intervals is True.
                [2] intervals is False and 'closed' is "left".
                [3] intervals is False and 'closed' is "right".
                [4] intervals is False and 'closed' is "both".
                [5] intervals is False and 'closed' is "neither".

            ends: list of tuple (see method signature) of:
                [0] value to pass as end.
                Other items define the end of slice of subset when end is [0] and:
                [1] intervals is True.
                [2] intervals is False and 'closed' is "left".
                [3] intervals is False and 'closed' is "right".
                [4] intervals is False and 'closed' is "both".
                [5] intervals is False and 'closed' is "neither".

            All other parameters will be passed thorugh to `trading_index`.
            """
            args_dates = (start_s, end_s, period)
            kwargs = dict(
                force=force,
                ignore_breaks=ignore_breaks,
                curtail_overlaps=curtail_overlaps,
            )
            for (start, slc_start, ssl, ssr, ssb, ssn), (
                end,
                slc_end,
                sel,
                ser,
                seb,
                sen,
            ) in itertools.product(starts, ends):
                args = (start, end, period)

                # verify for intervals index
                intervals = True
                index_dates = f(*args_dates, intervals, **kwargs)
                rtrn = f(*args, intervals=intervals, **kwargs)
                assert_index_equal(rtrn, index_dates[slc_start:slc_end])

                # verify for datetime index
                intervals = False
                closed = "left"
                rtrn = f(*args, intervals=intervals, closed=closed, **kwargs)
                index_dates = f(*args_dates, intervals, closed=closed, **kwargs)
                assert_index_equal(rtrn, index_dates[ssl:sel])

                closed = "right"
                rtrn = f(*args, intervals=intervals, closed=closed, **kwargs)
                index_dates = f(*args_dates, intervals, closed=closed, **kwargs)
                assert_index_equal(rtrn, index_dates[ssr:ser])

                closed = "both"
                rtrn = f(*args, intervals=intervals, closed=closed, **kwargs)
                index_dates = f(*args_dates, intervals, closed=closed, **kwargs)
                assert_index_equal(rtrn, index_dates[ssb:seb])

                closed = "neither"
                rtrn = f(*args, intervals=intervals, closed=closed, **kwargs)
                index_dates = f(*args_dates, intervals, closed=closed, **kwargs)
                assert_index_equal(rtrn, index_dates[ssn:sen])

        force, ignore_breaks = False, True

        period = one_min
        delta = period * 22

        starts = [
            (start_s, None, None, None, None, None),
            (start_s_open, None, None, None, None, None),
            (start_s_open + delta, 22, 22, 21, 22, 21),
            (start_s_open + delta - one_min, 21, 21, 20, 21, 20),
            (start_s_open + delta + one_min, 23, 23, 22, 23, 22),
        ]
        ends = [
            (end_s, None, None, None, None, None),
            (end_s_close, None, None, None, None, None),
            (end_s_close - delta, -22, -21, -22, -22, -21),
            (end_s_close - delta + one_min, -21, -20, -21, -21, -20),
            (end_s_close - delta - one_min, -23, -22, -23, -23, -22),
        ]

        assertions(starts, ends, period, force, ignore_breaks)

        period = pd.Timedelta(5, "min")
        delta = period * 2

        starts = [
            (start_s, None, None, None, None, None),
            (start_s_open, None, None, None, None, None),
            (start_s_open + delta, 2, 2, 1, 2, 1),
            (start_s_open + delta - one_min, 2, 2, 1, 2, 1),
            (start_s_open + delta + one_min, 3, 3, 2, 3, 2),
        ]
        ends = [
            (end_s, None, None, None, None, None),
            (end_s_close, None, None, None, None, None),
            (end_s_close - delta, -2, -1, -2, -2, -1),
            (end_s_close - delta + one_min, -2, -1, -2, -2, -1),
            (end_s_close - delta - one_min, -3, -2, -3, -3, -2),
        ]

        assertions(starts, ends, period, force, ignore_breaks)

        period = pd.Timedelta(1, "h")
        end_s_open = cal.session_open(end_s)

        # ignoring breaks...
        # assert assumption that end unaligned by 30mins
        assert (end_s_close - end_s_open) % period == pd.Timedelta(30, "min")

        end_s_aligned_post_close = end_s_close + pd.Timedelta(30, "min")
        end_s_break_end = cal.session_break_end(end_s)
        # assert assumption that pm session 3H duration
        assert end_s_close - end_s_break_end == pd.Timedelta(3, "h")

        starts = [
            (start_s, None, None, None, None, None),
            (start_s_open, None, None, None, None, None),
            (start_s_open + period, 1, 1, None, 1, None),
            (start_s_open + period - one_min, 1, 1, None, 1, None),
            (start_s_open + period + one_min, 2, 2, 1, 2, 1),
        ]
        ends = [
            (end_s, None, None, None, None, None),
            (end_s_aligned_post_close, None, None, None, None, None),
            (end_s_aligned_post_close + one_min, None, None, None, None, None),
            (end_s_aligned_post_close - one_min, -1, None, -1, -1, None),
            (end_s_close, -1, None, -1, -1, None),
            (end_s_aligned_post_close - period + one_min, -1, None, -1, -1, None),
            (end_s_aligned_post_close - period, -1, None, -1, -1, None),
            (end_s_aligned_post_close - period - one_min, -2, -1, -2, -2, -1),
            (end_s_break_end, -4, -3, -4, -4, -3),
            (end_s_break_end + pd.Timedelta(30, "min"), -3, -2, -3, -3, -2),
            (end_s_break_end + pd.Timedelta(29, "min"), -4, -3, -4, -4, -3),
            (end_s_break_end - pd.Timedelta(30, "min"), -4, -3, -4, -4, -3),
            (end_s_break_end - pd.Timedelta(31, "min"), -5, -4, -5, -5, -4),
        ]

        assertions(starts, ends, period, force, ignore_breaks)

        # verify effect of force
        starts = [
            (start_s, None, None, None, None, None),
            (start_s_open, None, None, None, None, None),
            (start_s_open + period, 1, 1, None, 1, None),
            (start_s_open + period - one_min, 1, 1, None, 1, None),
            (start_s_open + period + one_min, 2, 2, 1, 2, 1),
        ]
        ends = [
            (end_s, None, None, None, None, None),
            (end_s_aligned_post_close, None, None, None, None, None),
            (end_s_close, None, None, None, None, None),
            (end_s_close - one_min, -1, None, -1, -1, None),
            (end_s_close - pd.Timedelta(30, "min"), -1, None, -1, -1, None),
            (end_s_close - pd.Timedelta(31, "min"), -2, -1, -2, -2, -1),
            # break end as before...
            (end_s_break_end, -4, -3, -4, -4, -3),
            (end_s_break_end + pd.Timedelta(30, "min"), -3, -2, -3, -3, -2),
            (end_s_break_end + pd.Timedelta(29, "min"), -4, -3, -4, -4, -3),
            (end_s_break_end - pd.Timedelta(30, "min"), -4, -3, -4, -4, -3),
            (end_s_break_end - pd.Timedelta(31, "min"), -5, -4, -5, -5, -4),
        ]

        force = True
        assertions(starts, ends, period, force, ignore_breaks)

        # ACKNOWLEDGING BREAKS

        end_s_break_start = cal.session_break_start(end_s)
        # assert assumption that break start unaligned by 30mins
        assert (end_s_break_start - end_s_open) % period == pd.Timedelta(30, "min")

        starts = [
            (start_s, None, None, None, None, None),
            (start_s_open, None, None, None, None, None),
            (start_s_open + period, 1, 1, None, 1, None),
            (start_s_open + period - one_min, 1, 1, None, 1, None),
            (start_s_open + period + one_min, 2, 2, 1, 2, 1),
        ]
        ends = [
            (end_s, None, None, None, None, None),
            (end_s_aligned_post_close, None, None, None, None, None),
            (end_s_close, None, None, None, None, None),
            (end_s_close - one_min, -1, None, -1, -1, None),
            (end_s_close - period, -1, None, -1, -1, None),
            (end_s_close - period - one_min, -2, -1, -2, -2, -1),
            (end_s_break_end, -3, -2, -3, -3, -2),
            (end_s_break_end + one_min, -3, -2, -3, -3, -2),
            (end_s_break_end - one_min, -3, -3, -3, -4, -2),
            (end_s_break_start, -4, -3, -4, -5, -2),
            (end_s_break_start + pd.Timedelta(30, "min"), -3, -3, -3, -4, -2),
            (end_s_break_start + pd.Timedelta(29, "min"), -4, -3, -4, -5, -2),
            (end_s_break_start - pd.Timedelta(30, "min"), -4, -3, -4, -5, -2),
            (end_s_break_start - pd.Timedelta(31, "min"), -5, -4, -5, -6, -3),
        ]

        force, ignore_breaks = False, False
        # expected = expected_index(interval, force, ignore_breaks)
        assertions(starts, ends, period, force, ignore_breaks)

        # verifying effect of force when acknowledging breaks

        starts = [
            (start_s, None, None, None, None, None),
            (start_s_open, None, None, None, None, None),
            (start_s_open + period, 1, 1, None, 1, None),
            (start_s_open + period - one_min, 1, 1, None, 1, None),
            (start_s_open + period + one_min, 2, 2, 1, 2, 1),
        ]
        ends = [
            (end_s, None, None, None, None, None),
            (end_s_aligned_post_close, None, None, None, None, None),
            # end_s_close and end_s_break_end as before
            (end_s_close, None, None, None, None, None),
            (end_s_close - one_min, -1, None, -1, -1, None),
            (end_s_close - period, -1, None, -1, -1, None),
            (end_s_close - period - one_min, -2, -1, -2, -2, -1),
            (end_s_break_end, -3, -2, -3, -3, -2),
            (end_s_break_end + one_min, -3, -2, -3, -3, -2),
            (end_s_break_end - one_min, -3, -3, -3, -4, -2),
            # end_s_break_start affected by force
            (end_s_break_start, -3, -3, -3, -4, -2),
            (end_s_break_start - one_min, -4, -3, -4, -5, -2),
            (end_s_break_start - pd.Timedelta(30, "min"), -4, -3, -4, -5, -2),
            (end_s_break_start - pd.Timedelta(31, "min"), -5, -4, -5, -6, -3),
        ]

        force, ignore_breaks = True, False
        assertions(starts, ends, period, force, ignore_breaks)

    # PARSING TESTS

    def test_parsing_errors(self, cal_start_end):
        cal, start, end = cal_start_end
        error_msg = (
            "`period` cannot be greater than one day although received as"
            f" '{pd.Timedelta('2d')}'."
        )
        with pytest.raises(ValueError, match=error_msg):
            cal.trading_index(start, end, "2d", parse=False)

        error_msg = "If `intervals` is True then `closed` cannot be 'neither'."
        with pytest.raises(ValueError, match=re.escape(error_msg)):
            cal.trading_index(
                start, end, "20min", intervals=True, closed="neither", parse=False
            )

        error_msg = "If `intervals` is True then `closed` cannot be 'both'."
        with pytest.raises(ValueError, match=re.escape(error_msg)):
            cal.trading_index(
                start, end, "20min", intervals=True, closed="both", parse=False
            )

        # Verify raises error if period "1D" and start or end not passed as a date.
        start = pd.Timestamp("2018-05-01", tz=UTC)
        end = pd.Timestamp("2018-05-31")
        with pytest.raises(ValueError, match="a Date must be timezone naive"):
            cal.trading_index(start, end, "1D")

        start = pd.Timestamp("2018-05-01 00:01")
        with pytest.raises(
            ValueError, match="a Date must have a time component of 00:00"
        ):
            cal.trading_index(start, end, "1D")

        # verify raises wtih invalid values for `period`
        invalid_str = "X"
        error_msg = (
            f"`period` receieved as '{invalid_str}' although takes type"
            " 'pd.Timedelta' or a 'str' that is valid as a single input"
            " to 'pd.Timedelta'. Examples of valid input: pd.Timestamp('15T'),"
            " '15min', '15T', '1H', '4h', '1d', '5s', 500ms'."
        )
        with pytest.raises(ValueError, match=re.escape(error_msg)):
            cal.trading_index(start, end, invalid_str)

        invalid_value = pd.Timedelta(1441, "min")
        error_msg = re.escape(
            "`period` cannot be greater than one day although received as"
            f" '{invalid_value}'."
        )
        with pytest.raises(ValueError, match=error_msg):
            cal.trading_index(start, end, invalid_value)

        # verify raises wtih invalid values for `align` and `align_pm`
        error_msg = (
            f"`align` receieved as '{invalid_str}' although takes type"
            f" 'pd.Timedelta' or a 'str' that is valid as a single input"
            " to 'pd.Timedelta'. Examples of valid input: pd.Timestamp('5T'),"
            " '5min', '5T', pd.Timedelta('-5T'), '-5min', '-5T'."
        )
        with pytest.raises(ValueError, match=re.escape(error_msg)):
            cal.trading_index(start, end, "1h", align=invalid_str)

        error_msg = (
            f"`align_pm` receieved as '{invalid_str}' although takes type bool,"
            f" 'pd.Timedelta' or a 'str' that is valid as a single input"
            " to 'pd.Timedelta'. Examples of valid input: pd.Timestamp('5T'),"
            " '5min', '5T', pd.Timedelta('-5T'), '-5min', '-5T'."
        )
        with pytest.raises(ValueError, match=re.escape(error_msg)):
            cal.trading_index(start, end, "1h", align="5min", align_pm=invalid_str)

        invalid_values = [pd.Timedelta(7, "min"), pd.Timedelta(0), "0min"]
        for value in invalid_values:
            error_msg_end = (
                f" must be factor of 1H although received '{pd.Timedelta(value)}'."
            )
            with pytest.raises(ValueError, match=re.escape("`align`" + error_msg_end)):
                cal.trading_index(start, end, "1h", align=value)

            with pytest.raises(
                ValueError, match=re.escape("`align_pm`" + error_msg_end)
            ):
                cal.trading_index(start, end, "1h", align="5min", align_pm=value)

        invalid_minute_fractions = [3, pd.Timedelta(3600, "ms"), 3.6, "3s"]
        for value in invalid_minute_fractions:
            error_msg_end = (
                " cannot include a fraction of a minute although received "
                f"'{pd.Timedelta(value)}'."
            )
            with pytest.raises(ValueError, match=re.escape("`align`" + error_msg_end)):
                cal.trading_index(start, end, "1h", align=value)

            with pytest.raises(
                ValueError, match=re.escape("`align_pm`" + error_msg_end)
            ):
                cal.trading_index(start, end, "1h", align="5min", align_pm=value)



import pytest

from exchange_calendars.exchange_calendar_xmex import XMEXExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXMEXCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XMEXExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XMEX is open from 8:30AM to 3:00PM.
        yield 6.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2019-01-01",  # New Year's Day
            "2019-02-04",  # Constitution Day
            "2019-03-18",  # Juarez's Birthday
            "2019-04-18",  # Maundy Thursday
            "2019-04-19",  # Good Friday
            "2019-05-01",  # Labour Day
            "2019-09-16",  # Independence Day
            "2018-11-02",  # All Souls' Day
            "2019-11-18",  # Revolution Day
            "2019-12-12",  # Banking Holiday
            "2019-12-25",  # Christmas Day
            #
            "2006-11-02",  # first observance of regular holiday All Soul's Day.
            #
            # Rule changes.
            # Consitution Day. Prior to 2007 observed strictly on February 5th.
            "2004-02-05",  # falls Thursday, otherwise a trading day
            # from 2007 observed on first Monday of February
            "2008-02-04",  # 5th falls on a Tuesday, first Monday is 4th
            # Juarez's Birthday. Prior to 2007 observed strictly on March 21st.
            "2006-03-21",  # falls Tuesday, otherwise a trading day
            # from 2007 observed on third Monday of March.
            "2007-03-19",  # 21st falls on a Wednesday, third Monday is 19th.
            # Revolution Day. Prior to 2007 observed strictly on November 20th.
            "2003-11-20",  # falls Thursday, otherwise a trading day
            # from 2007 obseved on the third Monday of November.
            "2007-11-19",  # 20th falls on a Tuesday, third Monday is 19th.
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield ["2010-09-17"]  # Bicentennial Celebration.

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend and are not made up. Ensure surrounding
            # days are not holidays.
            # New Year's Day on a Sunday.
            "2016-12-30",
            "2017-01-02",
            # Constitution Day (February 5th) on a Sunday, prior to 2007.
            "2006-02-03",
            "2006-02-06",
            # Juarez's Birthday (March 21st) on a Sunday, prior to 2007.
            "2004-03-19",
            "2004-03-22",
            # Labour Day (May 1st) on a Sunday.
            "2016-04-29",
            "2016-05-02",
            # Independence Day (September 16th) on a Sunday.
            "2018-09-14",
            "2018-09-17",
            # All Souls' Day (November 2nd) on a Saturday.
            "2019-11-01",
            "2019-11-04",
            # Revolution Day (November 20th) on a Sunday, prior to 2007.
            "2005-11-18",
            "2005-11-21",
            # Banking Holiday (December 12th) on a Saturday.
            "2015-12-11",
            "2015-12-14",
            # Christmas on a Sunday.
            "2016-12-23",
            "2016-12-26",
            #
            "2005-11-02",  # All Soul's Day not observed until 2006.
            #
            # Rule changes.
            # Consitution Day. Prior to 2007 observed strictly on February 5th,
            # from 2007 observed on first Monday of February.
            "2006-02-06",  # fell on Sunday, not made up so Monday is not a holiday.
            "2004-02-02",  # prior to rule change first Monday is not a holiday.
            "2008-02-05",  # falls on a Tuesday, post rule change no longer a holiday.
            # Juarez's Birthday. Prior to 2007 observed strictly on March 21st,
            # from 2007 observed on third Monday of March.
            "2006-03-20",  # prior to rule change third Monday is not a holiday.
            "2007-03-21",  # falls on a Wednesday, post rule change no longer a holiday.
            # Revolution Day. Prior to 2007 observed strictly on November 20th,
            # from 2007 obseved on third Monday of November.
            "2003-11-17",  # prior to rule change third Monday is not a holiday.
            "2007-11-20",  # falls on a Tuesday, post rule change no longer a holiday.
        ]



import pytest

from exchange_calendars.exchange_calendar_xshg import XSHGExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase
from .test_utils import T


class TestXSHGCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XSHGExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # Shanghai stock exchange is open from 9:30 am to 3pm
        yield 5.5

    @pytest.fixture
    def start_bound(self):
        yield T("1990-12-03")

    @pytest.fixture
    def end_bound(self):
        yield T("2025-12-31")

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2017
            "2017-01-02",
            "2017-01-27",
            "2017-01-30",
            "2017-01-31",
            "2017-02-01",
            "2017-02-02",
            "2017-04-03",
            "2017-04-04",
            "2017-05-01",
            "2017-05-29",
            "2017-05-30",
            "2017-10-02",
            "2017-10-03",
            "2017-10-04",
            "2017-10-05",
            "2017-10-06",
            # 2020
            "2020-01-31",
            # 2022
            "2022-01-31",
            "2022-09-12",
            "2022-10-06",
            "2022-10-07",
            # 2023
            "2023-05-03",  # Part of Chinese Labor Day 2023
            "2023-06-23",  # Part of Dragon Boat Festival 2023
            # 2024
            "2024-01-01",
            # 2024 Chinese New Year
            "2024-02-12",
            "2024-02-13",
            "2024-02-14",
            "2024-02-15",
            "2024-02-16",
            "2024-04-04",  # Part of Qingming Festival 2024
            "2024-04-05",  # Part of Qingming Festival 2024
            "2024-05-01",  # Part of Chinese Labor Day 2024
            "2024-05-02",
            "2024-05-03",
            "2024-06-10",  # Part of Dragon Boat Festival 2024
            "2024-09-16",  # Part of Mid-Autumn Festival 2024
            "2024-09-17",  # Part of Mid-Autumn Festival 2024
            "2024-10-01",  # Part of National Day 2024
            "2024-10-02",  # Part of National Day 2024
            "2024-10-03",  # Part of National Day 2024
            "2024-10-04",  # Part of National Day 2024
        ]



import pytest
import pandas as pd
import pandas.testing as tm

from exchange_calendars.always_open import AlwaysOpenCalendar
from exchange_calendars.calendar_helpers import UTC
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestAlwaysOpenCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class", params=["left", "right"])
    def all_calendars_with_answers(self, request, calendars, answers):
        yield (calendars[request.param], answers[request.param])

    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield AlwaysOpenCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 24

    # Calendar-specific tests

    def test_open_every_day(self, default_calendar_with_answers):
        cal, ans = default_calendar_with_answers
        dates = pd.date_range(*ans.sessions_range)
        tm.assert_index_equal(cal.sessions, dates)

    def test_open_every_minute(self, calendars, answers, one_minute):
        cal, ans = calendars["left"], answers["left"]
        minutes = pd.date_range(*ans.trading_minutes_range, freq="min", tz=UTC)
        tm.assert_index_equal(cal.minutes, minutes)

        cal = calendars["right"]
        minutes += one_minute
        tm.assert_index_equal(cal.minutes, minutes)



import pytest

import pandas as pd

from exchange_calendars.exchange_calendar_xjse import XJSEExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXJSECalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XJSEExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The JSE is open from 09:00 to 17:00.
        yield 8

    def test_no_weekend_sessions(self, default_calendar):
        bv = default_calendar.sessions.weekday.isin((5, 6))
        assert default_calendar.sessions[bv].empty

    @pytest.mark.parametrize(
        "year, holidays",
        [
            (
                2019,
                [
                    "2019-01-01",  # New Year's Day
                    "2019-03-21",  # Human Rights Day
                    "2019-04-19",  # Good Friday
                    "2019-04-22",  # Family Day
                    "2019-04-27",  # Freedom Day (falls on Saturday, not made up)
                    "2019-05-01",  # Workers' Day
                    "2019-05-08",  # Election Day (ad-hoc)
                    "2019-06-16",  # Youth Day
                    "2019-06-17",  # Youth Day (Monday make-up)
                    "2019-08-09",  # National Women's Day
                    "2019-09-24",  # Heritage Day
                    "2019-12-16",  # Day of Reconciliation
                    "2019-12-25",  # Christmas
                    "2019-12-26",  # Day of Goodwill
                ],
            ),
            (
                2018,
                [
                    "2018-01-01",  # New Year's Day
                    "2018-03-21",  # Human Rights Day
                    "2018-03-30",  # Good Friday
                    "2018-04-02",  # Family Day
                    "2018-04-27",  # Freedom Day
                    "2018-05-01",  # Workers' Day
                    "2018-06-16",  # Youth Day (falls on Saturday, not made up)
                    "2018-08-09",  # National Women's Day
                    "2018-09-24",  # Heritage Day
                    "2018-12-16",  # Day of Reconciliation
                    "2018-12-17",  # Day of Reconciliation (Monday make-up)
                    "2018-12-25",  # Christmas
                    "2018-12-26",  # Day of Goodwill
                ],
            ),
            (
                2016,
                [
                    "2016-01-01",  # New Year's Day
                    "2016-03-21",  # Human Rights Day
                    "2016-03-25",  # Good Friday
                    "2016-03-28",  # Family Day
                    "2016-04-27",  # Freedom Day
                    "2016-05-01",  # Workers' Day
                    "2016-05-02",  # Workers' Day (Monday make-up)
                    "2016-06-16",  # Youth Day
                    "2016-08-03",  # Election Day
                    "2016-08-09",  # National Women's Day
                    "2016-09-24",  # Heritage Day (falls on Saturday, not made up)
                    "2016-12-16",  # Day of Reconciliation
                    "2016-12-25",  # Christmas
                    "2016-12-26",  # Christmas (Monday make-up)
                    "2016-12-27",  # Day of Goodwill (Ad-hoc make-up observance)
                ],
            ),
        ],
    )
    def test_holidays_in_year(self, default_calendar, year, holidays):
        cal = default_calendar
        days = pd.date_range(start=f"{year}-01-01", end=f"{year}-12-31", freq="B")
        days = days.strftime("%Y-%m-%d")

        for holiday in holidays:
            # Sanity check
            assert holiday in days or pd.Timestamp(holiday).weekday() in (5, 6)

        for day in days:
            if day in holidays:
                assert not cal.is_session(day)
            else:
                assert cal.is_session(day)



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xwbo import XWBOExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXWBOCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XWBOExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XWBO is open from 9:00 am to 5:30 pm.
        yield 8.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2014 +
            "2014-01-01",  # New Year's Day
            "2014-01-06",  # Epiphany
            "2014-04-18",  # Good Friday
            "2014-04-21",  # Easter Monday
            "2014-05-01",  # Labour Day
            "2014-05-29",  # Ascension Day
            "2014-06-09",  # Whit Monday
            "2014-06-19",  # Corpus Christi
            "2014-08-15",  # Assumption Day
            "2014-10-26",  # National Day (Weekend)
            "2015-10-26",  # National Day (Weekday)
            "2013-11-01",  # All Saints Day (Weekday)
            "2014-11-01",  # All Saints Day (Weekend)
            "2014-12-08",  # Immaculate Conception
            "2014-12-24",  # Christmas Eve
            "2014-12-25",  # Christmas Day
            "2014-12-26",  # St. Stephen's Day
            "2014-12-31",  # New Year's Eve
            #
            # 2019
            "2019-01-01",  # New Year's Day
            "2019-04-19",  # Good Friday
            "2019-04-22",  # Easter Monday
            "2019-05-01",  # Labour Day
            "2019-06-10",  # Whit Monday
            "2019-10-26",  # National Day (Weekend)
            "2019-12-24",  # Christmas Eve
            "2019-12-25",  # Christmas Day
            "2019-12-26",  # St. Stephen's Day
            "2019-12-31",  # New Year's Eve
            #
            # Prior to 2016, when New Year's Eve fell on the weekend, it was
            # observed on the preceding Friday.
            "2011-12-30",
            "2006-12-29",
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield []

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend and are not made up. Ensure surrounding
            # days are not holidays.
            # Epiphany (January 6th) on a Saturday.
            "2018-01-05",
            "2018-01-08",
            # Assumption Day (August 15th) on a Saturday.
            "2015-08-14",
            "2015-08-17",
            # Labour Day (May 1st) on a Saturday.
            "2010-04-30",
            "2010-05-03",
            # National Day (October 26th) on a Sunday.
            "2014-10-24",
            "2014-10-27",
            # All Saints Day (November 1st) on a Sunday.
            "2015-10-30",
            "2015-11-02",
            # Immaculate Conception (December 8th) on a Saturday.
            "2018-12-07",
            "2018-12-10",
            # Christmas Eve on a Saturday and Christmas on a Sunday. This means
            # that the market should be open on the previous Friday, closed on
            # Monday for St. Stephen's Day, and open again on Tuesday.
            "2011-12-23",
            "2011-12-27",
            #
            # New Year's Eve not made up since 2016. Ensure day on which previously
            # made up (prior Friday) is not a holiday.
            "2016-12-30",
            "2017-12-29",
            # Whit Monday was a trading day in 2023
            "2023-05-29",
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            # Last trading day preceeding New Year's Eve
            "2017-12-29",
            "2018-12-28",
            "2019-12-30",
            "2020-12-30",
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=14, minutes=15)



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xdus import XDUSExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXDUSCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XDUSExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # open from 8am to 10pm
        yield 14

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2012
            # New Year's Day fell on a Sunday, so it is not a holiday this year
            "2012-04-06",  # Good Friday
            "2012-04-09",  # Easter Monday
            "2012-05-01",  # Labour Day
            # Whit Monday was observed in 2007, then 2015 and after.
            # German Unity Day started being celebrated in 2014
            "2012-12-24",  # Christmas Eve
            "2012-12-25",  # Christmas
            "2012-12-26",  # Boxing Day
            "2012-12-31",  # New Year's Eve
            #
            # Whit Monday
            "2015-05-25",  # regularly observed from 2015
            "2016-05-16",
            "2021-05-24"
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            "2007-05-28",  # Whit Monday observed as a one-off in 2007
            "2017-10-31",  # Reformation Day observed as a one-off in 2017
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            "2012-12-28",  # Last working day of 2012
            #
            # Whit Monday
            "2006-06-05",  # not observed prior to 2007 (observed in 2007)
            "2008-05-12",  # and not observed from 2008 through 2014
            "2022-06-06",  # not observed in 2022
            #
            # Reformation Day observed only in 2017, ensure not a holiday
            # in surrounding years.
            "2016-10-31",
            "2018-10-31",
            # German Unity Day was not observed in 2022
            "2022-10-03"
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2011-12-30",  # NYE on Sat, so Fri is a half day
            "2012-12-28",  # NYE on Mon, so preceding Fri is a half day
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=14)



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xtse import XTSEExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXTSECalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XTSEExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 6.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2012-01-02",  # New Year's Day 1st Jan fell on a Sunday, made up Monday.
            "2012-02-20",  # Family Day
            "2012-04-06",  # Good Friday
            "2012-05-21",  # Victoria Day
            "2012-07-02",  # Canada Day
            "2012-08-06",  # Civic Holiday
            "2012-09-03",  # Labour Day
            "2012-10-08",  # Thanksgiving
            "2012-12-25",  # Christmas
            "2012-12-26",  # Boxing Day
            #
            "2015-05-18",  # Victoria Day observed Monday before 25th, never Mon 25th.
            #
            # Holidays falling on a weekend and made up.
            "2010-12-27",  # Christmas Day 25th fell on Saturday, made up Monday.
            "2010-12-28",  # Boxing Day 26th fell on Sunday, made up Tuesday.
            "2015-12-28",  # Boxing Day 26th fell on Saturday, made up Monday.
            "2016-12-26",  # Christmas Day 26th fell on Monday, make sure holiday.
            "2016-12-27",  # Christmas Day 25th fell on Sunday, made up Tuesday.)
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            "2001-09-11",  # 9/11
            "2001-09-12",  # 9/11
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Make sure made up holidays do not extend beyond last made up date.
            "2010-12-29",  # First day (Wed) post Boxing Day that is not a holiday.
            "2016-12-28",  # First day (Wed) post Boxing Day that is not a holiday.
            #
            "2015-05-25",  # Victoria Day observed Monday before 25th, never Mon 25th.
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2010-12-24",  # Christmas Eve, first year that observed early close.
            "2012-12-24",  # Christmas Eve
            "2013-12-24",  # Christmas Eve
            "2014-12-24",  # Christmas Eve
            "2015-12-24",  # Christmas Eve
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(13, "h")

    @pytest.fixture
    def non_early_closes_sample(self):
        yield [
            "2009-12-24",  # Christmas Eve, last year not observing early close.
        ]

    @pytest.fixture
    def non_early_closes_sample_time(self):
        yield pd.Timedelta(16, "h")



import pytest

from exchange_calendars.exchange_calendar_xtai import XTAIExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXTAICalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XTAIExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XTAI is open from 9:00AM to 1:30PM
        yield 4.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2019
            "2019-01-01",  # New Year's Day
            "2019-02-04",  # Chinese New Year's Eve
            "2019-02-05",  # Chinese New Year
            "2019-02-28",  # Peace Memorial Day
            "2019-04-04",  # Women and Children's Day
            "2019-04-05",  # Tomb Sweeping Day
            "2019-05-01",  # Labour Day
            "2019-06-07",  # Dragon Boat Festival
            "2019-09-13",  # Mid-Autumn Festival
            "2019-10-10",  # National Day
            #
            # Holidays falling on a Sat/Sun are made up on Fri/Mon respectively.
            "2017-01-02",  # New Year's Day on Sunday, Jan 1st.
            "2017-01-27",  # Chinese New Year on Saturday, Jan 28th.
            "2016-02-29",  # Peace Memorial Day on Sunday, Feb 28th.
            "2015-04-03",  # Women and Children's Day on Saturday, Apr 4th.
            "2015-04-06",  # Tomb Sweeping Day on Sunday, Apr 5th.
            "2016-05-02",  # Labour Day on Sunday, May 1st.
            "2015-06-19",  # Dragon Boat Festival on Saturday, Jun 20th.
            "2015-09-28",  # Mid-Autumn Festival on Sunday, Sep 27th.
            "2015-10-09",  # National Day on Saturday, Oct 10th.
            #
            # "bridge days" where a Monday or Friday is made into a holiday
            # to fill gap between weekend and a holiday on Tuesday or Thursday.
            "2015-01-02",  # New Year's Day on Thursday, Jan 1st.
            "2017-02-27",  # Peace Memorial day on Tuesday, Feb 28th.
            "2017-04-03",  # Women and Children's Day on Tuesday, Apr 4th.
            "2019-10-11",  # National Day on Thursday, Oct 10th.
            "2018-04-06",  # Tomb Sweeping Day on Thursday, Apr 5th.
            "2017-05-29",  # Dragon Boat Festival on Tuesday, May 30th.
            "2016-09-16",  # Mid-Autumn Festival on Thursday, Sep 15th.
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays falling on a Sat/Sun are made up on Fri/Mon respectively,
            # make sure made up in correct direction (e.g. if fell on Sat, make sure
            # Mon not a holiday / if fell on Sun, make sure Fri not a holiday).
            "2016-12-30",  # New Year's Day on Sunday, Jan 1st.
            "2016-02-26",  # Peace Memorial Day on Sunday, Feb 28th.
            "2014-04-07",  # Tomb Sweeping Day on Saturday, Apr 5th.
            "2016-04-29",  # Labour Day on Sunday, May 1st.
            "2015-06-22",  # Dragon Boat Festival on Saturday, Jun 20th.
            "2015-09-25",  # Mid-Autumn Festival on Sunday, Sep 27th.
            "2015-10-12",  # National Day on Saturday, Oct 10th.
            "2024-09-16",  # Mid-Autumn Festival on Tuesday, Sep 17th.
            "2024-10-11",  # National Day on Thursday, Oct 10th.
            "2030-09-13",  # Mid-Autumn Festival on Thursday, Sep 12th.
        ]



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xnys import XNYSExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXNYSCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XNYSExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 6.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2012
            "2012-01-02",
            "2012-01-16",
            "2012-02-20",
            "2012-04-06",
            "2012-05-28",
            "2012-07-04",
            "2012-09-03",
            "2012-11-22",
            "2012-12-25",
            #
            # Thanksgiving. Fourth Thursday of November
            "2005-11-24",  # November has 4 Thursdays
            "2006-11-23",  # November has 5 Thursdays
            # Holidays falling on a weekend and subsequently made up on Monday.
            "2012-01-02",  # New Year's Day fell on Sunday.
            # Juneteenth 2022-2024
            "2022-06-20",
            "2023-06-19",
            "2024-06-19",
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        # NOTE: any date that is earlier than the first date of xnys.csv will NOT be
        # tested!
        yield [
            "2001-09-11",  # 9/11
            "2001-09-12",  # 9/11
            "2001-09-13",  # 9/11
            "2001-09-14",  # 9/11
            "2012-10-29",  # Hurricane Sandy
            "2012-10-30",  # Hurricane Sandy
            "1985-09-27",  # Hurricane Gloria
            "1977-07-14",  # New York Blackout
            "1972-11-07",  # pre-1980 Presidential Election Days
            "1976-11-02",  # pre-1980 Presidential Election Days
            "1980-11-04",  # pre-1980 Presidential Election Days
            "2018-12-05",  # George H.W. - 12/5/2018
            "2007-01-02",  # Gerald Ford
            "2004-06-11",  # Ronald Reagan
            "1994-04-27",  # Richard Nixon
            "1973-01-25",  # Lyndon B. Johnson
            "1972-12-28",  # Harry S. Truman
            "2025-01-09",  # Jimmy Carter
            # National Days of Mourning
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Thanksgiving. Fourth Thursday of November
            "2006-11-30",  # November has 5 Thursdays, last Thursday is not a holiday.
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            # 2012
            "2012-07-03",
            "2012-11-23",  # Day after Thanksgiving
            "2012-12-24",  # Christmas Eve
            #
            "2013-11-29",  # Day after Thanksgiving, 4 Thursdays, 5 Fridays in month.
            #
            # Until 2013, if Independence Day fell on a Thursday, the market closed
            # early on the Friday. Since 2013 the early close is on the Wednesday.
            "2002-07-05",  # Friday after an Independence Day that fell on a Thursday
            "2013-07-03",  # Wednesday before an Independence Day that fell on Thursday
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(13, "h")

    @pytest.fixture
    def non_early_closes_sample(self):
        yield [
            # Until 2013, if Independence Day fell on a Thursday, the market closed
            # early on the Friday. Since 2013 the early close is on the Wednesday.
            "2002-07-03",  # Wednesday before an Independence Day that fell on Thursday
            "2013-07-05",  # Friday after an Independence Day that fell on Thursday
        ]

    @pytest.fixture
    def non_early_closes_sample_time(self):
        yield pd.Timedelta(16, "h")



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xdub import XDUBExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXDUBCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XDUBExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XDUB is open from 8:00 am to 4:28 pm.
        yield 8.467

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2018
            "2018-01-01",  # New Year's Day
            "2018-03-30",  # Good Friday
            "2018-04-02",  # Easter Monday
            "2018-05-07",  # May Bank Holiday
            "2018-06-04",  # June Bank Holiday
            "2018-12-25",  # Christmas Day
            "2018-12-26",  # Boxing Day
            #
            # Holidays falling on a weekend are made up on what would otherwise be
            # the next trading day.
            # In 2017 New Year's Day fell on a Sunday, so the following Monday
            # should be a holiday.
            "2017-01-02",
            # In 2010 Christmas fell on a Saturday, meaning Boxing Day fell on
            # a Sunday. The following Monday and Tuesday should both be
            # holidays.
            "2010-12-27",
            "2010-12-28",
            # In 2016 Christmas fell on a Sunday, but again the following
            # Monday and Tuesday should both be holidays.
            "2016-12-26",
            "2016-12-27",
            #
            # Regular holiday's that later ceased to be observed.
            "1996-03-18",  # St. Patrick's Day Observed
            "2000-03-17",  # St. Patrick's Day
            "2009-05-01",  # Labour Day
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield ["2018-03-02"]  # March 2, 2018 was closed due to sever weather.

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Regular holidays that earlier ceased to be observed.
            "2001-03-19",  # St. Patrick's Day Observed
            "2003-03-17",  # St. Patrick's Day
            "2012-05-01",  # Labour Day
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2010-12-24",  # Christmas Eve on a weekday.
            "2018-12-24",  # Christmas Eve on a weekday.
            "2017-12-22",  # Last trading day prior to Christmas Eve.
            "2010-12-31",  # New Year's Eve on a weekday.
            "2018-12-31",  # New Year's Eve on a weekday.
            "2017-12-29",  # Last trading day prior to New Year's Eve.
            "2018-03-01",  # Severe weather, closed early.
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=12, minutes=28)

    @pytest.fixture
    def non_early_closes_sample(self):
        yield [
            # Prior to 2010 Christmas Eve and New Year's Eve were full days
            "2009-12-24",
            "2009-12-31",
            #
            # Just a normal day...
            "2017-03-01",
        ]

    @pytest.fixture
    def non_early_closes_sample_time(self):
        yield pd.Timedelta(hours=16, minutes=28)



# -*- coding: utf-8 -*-
#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time, timedelta
from itertools import chain
from zoneinfo import ZoneInfo

import numpy as np
import pandas as pd
import toolz
from pandas.tseries.holiday import EasterMonday, GoodFriday, Holiday, sunday_to_monday
from pandas.tseries.offsets import LastWeekOfMonth, WeekOfMonth

from .common_holidays import (
    boxing_day,
    christmas,
    christmas_eve,
    new_years_day,
    new_years_eve,
    weekend_christmas,
)
from .lunisolar_holidays import (
    chinese_buddhas_birthday_dates,
    chinese_lunar_new_year_dates,
    double_ninth_festival_dates,
    dragon_boat_festival_dates,
    mid_autumn_festival_dates,
    qingming_festival_dates,
)
from .exchange_calendar import (
    FRIDAY,
    MONDAY,
    SATURDAY,
    SUNDAY,
    THURSDAY,
    TUESDAY,
    WEDNESDAY,
    HolidayCalendar,
)
from .precomputed_exchange_calendar import PrecomputedExchangeCalendar
from .utils.pandas_utils import vectorized_sunday_to_monday

# Useful resources for making changes to this file:
# # /etc/lunisolar
# http://www.math.nus.edu.sg/aslaksen/calendar/cal.pdf
# https://www.hko.gov.hk/gts/time/calendarinfo.htm
#   - the almanacs on this page are also useful

weekdays = (MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY)


def process_queen_birthday(dt):
    # before 1983
    if dt.year in [1974, 1981]:
        return dt + pd.DateOffset(weekday=6)
    elif dt.year < 1983:
        return sunday_to_monday(dt)
    # after 1983
    wom = WeekOfMonth(week=2, weekday=0)
    if dt.year in [1983, 1988, 1993, 1994]:
        wom = WeekOfMonth(week=1, weekday=0)
    if dt.year in [1985]:
        wom = WeekOfMonth(week=3, weekday=0)
    return dt + wom


LabourDay = Holiday(
    name="Labour Day",  # 劳动节
    month=5,
    day=1,
    observance=sunday_to_monday,
    start_date=pd.Timestamp("1999-05-01"),
)

HKRegionEstablishmentDay = Holiday(
    name="Hong Kong Special Region Establishment Day",
    month=7,
    day=1,
    observance=sunday_to_monday,
    start_date=pd.Timestamp("1997-07-01"),
)

NationalDay = Holiday(
    name="National Day",
    month=10,
    day=1,
    observance=sunday_to_monday,
    start_date=pd.Timestamp("1997-07-01"),
)

QueenBirthday = Holiday(
    name="Queen's Birthday",  # 英女王生日 6月
    month=6,
    day=10,
    observance=process_queen_birthday,
    start_date=pd.Timestamp("1983-01-01"),
    end_date=pd.Timestamp("1997-06-01"),
)

QueenBirthday2 = Holiday(
    name="Queen's Birthday",  # 英女王生日 4月
    month=4,
    day=21,
    observance=process_queen_birthday,
    start_date=pd.Timestamp("1926-04-21"),
    end_date=pd.Timestamp("1983-01-01"),
)

CommemoratingAlliedVictory = Holiday(
    name="Commemorating the allied victory",  # 重光纪念日 8月最后一个星期一
    month=8,
    day=20,
    offset=LastWeekOfMonth(weekday=0),
    start_date=pd.Timestamp("1945-08-30"),
    end_date=pd.Timestamp("1997-07-01"),
)

IDontKnow = Holiday(
    name="I dont know these days, please tell me",  # 8月第一个星期一
    month=7,
    day=31,
    offset=WeekOfMonth(week=0, weekday=0),
    start_date=pd.Timestamp("1960-08-01"),
    end_date=pd.Timestamp("1983-01-01"),
)


day_after_mid_autumn_festival_dates = mid_autumn_festival_dates + timedelta(1)

HKAdhocClosures = [
    # I dont know these days
    pd.Timestamp("1970-07-01"),
    pd.Timestamp("1971-07-01"),
    pd.Timestamp("1973-07-02"),
    pd.Timestamp("1974-07-01"),
    pd.Timestamp("1975-07-01"),
    pd.Timestamp("1976-07-01"),
    pd.Timestamp("1977-07-01"),
    pd.Timestamp("1979-07-02"),
    pd.Timestamp("1980-07-01"),
    pd.Timestamp("1981-07-01"),
    pd.Timestamp("1982-07-01"),
    pd.Timestamp("1971-03-22"),
    pd.Timestamp("1971-12-06"),
    pd.Timestamp("1971-12-20"),
    pd.Timestamp("1975-07-28"),
    pd.Timestamp("1985-07-29"),
    # Weather related closures
    pd.Timestamp("1970-07-16"),  # 台风Ruby7003
    pd.Timestamp("1970-09-14"),  # 台风Georgia7011
    pd.Timestamp("1971-07-22"),  # 台风Lucy7114
    pd.Timestamp("1971-08-31"),  # 重光纪念日?
    pd.Timestamp("1973-04-16"),  # 股灾休市?
    pd.Timestamp("1973-07-17"),  # 台风Dot7304
    pd.Timestamp("1974-04-25"),  # 英国女王生日
    pd.Timestamp("1975-10-14"),  # 台风Elsie7514
    pd.Timestamp("1978-07-26"),  # 台风Agnes7807
    pd.Timestamp("1978-07-27"),
    pd.Timestamp("1979-01-26"),  # 春节补假
    pd.Timestamp("1979-08-02"),  # 台风Hope7908
    pd.Timestamp("1980-05-21"),  # 台风Georgia8004
    pd.Timestamp("1980-07-22"),  # 台风Joy8007
    pd.Timestamp("1981-04-27"),  # 英国女王生日
    pd.Timestamp("1981-07-06"),  # 台风Lynn8106
    pd.Timestamp("1981-07-07"),
    pd.Timestamp("1981-07-29"),  # 查理斯王子与戴安娜婚礼
    pd.Timestamp("1983-09-09"),  # 台风Ellen8309
    pd.Timestamp("1985-06-24"),  # 台风Hal8504
    pd.Timestamp("1986-04-01"),  # 复活节星期一翌日
    pd.Timestamp("1986-10-22"),  # 英女王伊丽莎白二世访港
    pd.Timestamp("1987-10-20"),  # 黑色星期一后,休市4天
    pd.Timestamp("1987-10-21"),
    pd.Timestamp("1987-10-22"),
    pd.Timestamp("1987-10-23"),
    pd.Timestamp("1988-04-05"),  # 清明节翌日
    # Timestamp('1988-06-13'),  # 英国女王生日
    pd.Timestamp("1991-06-18"),  # 英国女王生日翌日
    pd.Timestamp("1992-07-22"),  # 台风Cary9207
    # pd.Timestamp('1993-06-14'),  # 英国女王生日
    pd.Timestamp("1993-09-17"),  # 台风Becky9316
    pd.Timestamp("1994-06-14"),  # 英国女王生日翌日,端午节翌日
    pd.Timestamp("1997-06-30"),  # 英国女王生日
    pd.Timestamp("1997-07-02"),  # 香港回归纪念日翌日
    pd.Timestamp("1997-08-18"),  # 抗战胜利纪念日
    pd.Timestamp("1997-10-02"),  # 国庆节翌日
    pd.Timestamp("1998-08-17"),  # 抗战胜利纪念日
    pd.Timestamp("1998-10-02"),  # 国庆节翌日
    pd.Timestamp("1999-04-06"),  # 清明节翌日
    pd.Timestamp("1999-09-16"),  # 台风约克
    pd.Timestamp("1999-12-31"),  # 千年虫
    pd.Timestamp("2001-07-06"),  # 台风尤特0104
    pd.Timestamp("2001-07-25"),  # 台风玉兔0107
    # pd.Timestamp(2008-06-25'),  # 台风风神0806,上午休市
    pd.Timestamp("2008-08-06"),  # 台风北冕0809
    pd.Timestamp("2008-08-22"),  # 台风鹦鹉0810
    # pd.Timestamp(2009-09-15'),  # 台风巨爵0915,上午休市
    pd.Timestamp("2010-04-06"),  # 清明节翌日
    pd.Timestamp("2011-09-29"),  # 台风纳沙1117
    # pd.Timestamp(2012-07-24'),  # 台风韦森特1208,上午休市
    pd.Timestamp("2012-10-02"),  # 中秋节补假
    # pd.Timestamp(2013-05-22'),  # 暴雨,上午休市
    pd.Timestamp("2013-08-14"),  # 台风尤特1311
    # pd.Timestamp(2013-09-23'),  # 台风天兔1319,上午休市
    # pd.Timestamp(2014-09-16'),  # 台风海鸥1415,上午休市
    pd.Timestamp("2015-04-07"),  # 复活节+清明节补假
    # pd.Timestamp(2015-07-09'),  # 台风莲花1520,期货夜盘休市
    pd.Timestamp("2015-09-03"),  # 抗战70周年纪念
    # pd.Timestamp(2016-08-01'),  # 台风妮妲1604,期货夜盘20:55收市
    pd.Timestamp("2016-08-02"),  # 台风妮妲1604
    pd.Timestamp("2016-10-21"),  # 台风海马1622
    # pd.Timestamp(2017-06-12'),  # 台风苗柏1702,期货夜盘17:35休市
    pd.Timestamp("2017-08-23"),  # 台风天鸽1713
    pd.Timestamp("2020-10-13"),  # 台风浪卡2016
    pd.Timestamp(
        "2021-10-13"
    ),  # https://www.hkex.com.hk/News/Market-Communications/2021/2110132news?sc_lang=en
    # In 2022, the mid-autumn festival falls on a Saturday (2022-09-10),
    #  which means the day after is a Sunday. In 2022, HKSE had a holiday
    #  on the Monday (2022-09-12). In the past they don't seem to have followed
    #  this pattern. We'll have to wait and see before we generalise this into a rule.
    pd.Timestamp("2022-09-12"),
    pd.Timestamp("2023-07-17"),  # 8号台风泰利, 全天休市
    pd.Timestamp("2024-09-06"),  # 八號颱風，全天休市
]


def boxing_day_obs(dt):
    if dt.weekday in (MONDAY, TUESDAY):
        return dt + pd.Timedelta(days=1)
    return dt


class XHKGExchangeCalendar(PrecomputedExchangeCalendar):
    """
    Exchange calendar for the Hong Kong Stock Exchange (XHKG).

    Open Time: 9:30 AM, Asia/Hong_Kong
    Lunch Break: 12:00 PM - 1:00 PM Asia/Hong_Kong
    Close Time: 4:00 PM, Asia/Hong_Kong

    Regularly-Observed Holidays:
    - New Years Day (observed on monday when Jan 1 is a Sunday)
    - Lunar New Year and the following 2 days. If the 3rd day of the lunar year
      is a Sunday, then the next Monday is a holiday.
    - Ching Ming Festival
    - Good Friday
    - Easter Monday
    - Buddhas Birthday
    - Dragon Boat Festival
    - Chinese National Day (observed on monday when Oct 1 is a Sunday)
    - Day Following Mid-Autumn Festival
    - Chung Yeung Festival
    - Christmas (observed on nearest weekday to December 25)
    - Day after Christmas is observed

    Regularly-Observed Early Closes:
    - Lunar New Year's Eve
    - Christmas Eve
    - New Year's Eve


    Additional Irregularities:
    - Closes frequently for severe weather.

    See https://www.hkex.com.hk/Services/Trading-hours-and-Severe-Weather-Arrangements/Trading-Hours/Securities-Market?sc_lang=en # noqa
    """

    name = "XHKG"
    tz = ZoneInfo("Asia/Hong_Kong")

    open_times = (
        (None, time(10)),
        (pd.Timestamp("2011-03-07"), time(9, 30)),
    )
    break_start_times = ((None, time(12, 0)),)
    break_end_times = ((None, time(13, 0)),)
    close_times = ((None, time(16)),)
    regular_early_close_times = (
        (None, time(12, 30)),
        (pd.Timestamp("2011-03-07"), time(12, 00)),
    )

    @classmethod
    def precomputed_holidays(cls):
        lunisolar_holidays = (
            chinese_buddhas_birthday_dates,
            chinese_lunar_new_year_dates,
            day_after_mid_autumn_festival_dates,
            double_ninth_festival_dates,
            dragon_boat_festival_dates,
            qingming_festival_dates,
        )
        return lunisolar_holidays

    @classmethod
    def _earliest_precomputed_year(cls) -> int:
        return max(map(np.min, cls.precomputed_holidays())).year

    @classmethod
    def _latest_precomputed_year(cls) -> int:
        return min(map(np.max, cls.precomputed_holidays())).year

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                new_years_day(observance=sunday_to_monday),
                GoodFriday,
                EasterMonday,
                LabourDay,
                HKRegionEstablishmentDay,
                CommemoratingAlliedVictory,
                IDontKnow,
                NationalDay,
                QueenBirthday,
                QueenBirthday2,
                christmas(),
                weekend_christmas(),
                boxing_day(observance=boxing_day_obs),
            ]
        )

    @property
    def adhoc_holidays(self):
        # overrides as inherited from PrecomputedExchangeCalendar
        lunar_new_years_eve = (chinese_lunar_new_year_dates - pd.Timedelta(days=1))[
            (chinese_lunar_new_year_dates.weekday == SATURDAY)
            & (chinese_lunar_new_year_dates.year < 2013)
        ]

        lunar_new_year_2 = chinese_lunar_new_year_dates + pd.Timedelta(days=1)
        lunar_new_year_3 = chinese_lunar_new_year_dates + pd.Timedelta(days=2)
        lunar_new_year_4 = (chinese_lunar_new_year_dates + pd.Timedelta(days=3))[
            # According to the new arrangement under the General Holidays and
            # Employment Legislation (Substitution of Holidays)(Amendment)
            # Ordinance 2011, when either Lunar New Year's Day, the second day
            # of Lunar New Year or the third day of Lunar New Year falls on a
            # Sunday, the fourth day of Lunar New Year is designated as a
            # statutory and general holiday in substitution.
            (
                (chinese_lunar_new_year_dates.weekday == SUNDAY)
                | (lunar_new_year_2.weekday == SUNDAY)
                | (lunar_new_year_3.weekday == SUNDAY)
            )
            & (chinese_lunar_new_year_dates.year >= 2013)
        ]

        qingming_festival = vectorized_sunday_to_monday(
            qingming_festival_dates,
        ).values.copy()  # copy so that array is writeable
        years = qingming_festival.astype("M8[Y]")
        easter_monday = EasterMonday.dates(years[0], years[-1] + 1)
        # qingming gets observed one day later if easter monday is on the same
        # day
        qingming_festival[qingming_festival == easter_monday] += np.timedelta64(1, "D")

        # if the day after the mid autumn festival is October first, which
        # conflicts with national day, then national day is observed on the
        # second, though we don't encode that in the regular holiday, so
        # instead we pretend that the mid autumn festival would be delayed.
        mid_autumn_festival = day_after_mid_autumn_festival_dates.values.copy()
        mid_autumn_festival[
            (day_after_mid_autumn_festival_dates.month == 10)
            & (day_after_mid_autumn_festival_dates.day == 1)
        ] += np.timedelta64(1, "D")

        return list(
            chain(
                lunar_new_years_eve,
                chinese_lunar_new_year_dates,
                lunar_new_year_2,
                lunar_new_year_3,
                lunar_new_year_4,
                qingming_festival,
                vectorized_sunday_to_monday(chinese_buddhas_birthday_dates),
                vectorized_sunday_to_monday(dragon_boat_festival_dates),
                mid_autumn_festival,
                vectorized_sunday_to_monday(double_ninth_festival_dates),
                HKAdhocClosures,
            )
        )

    @property
    def special_closes(self):
        return [
            # HK changed their early close time
            (
                time(12, 30),
                HolidayCalendar(
                    [
                        new_years_eve(
                            # Market was close for Y2K instead of closing early
                            end_date=pd.Timestamp("1999-12-01"),
                            days_of_week=weekdays,
                        ),
                        new_years_eve(
                            start_date=pd.Timestamp("2000-12-01"),
                            end_date=pd.Timestamp("2011-03-07"),
                            days_of_week=weekdays,
                        ),
                        christmas_eve(
                            end_date=pd.Timestamp("2011-03-07"), days_of_week=weekdays
                        ),
                    ]
                ),
            ),
            (
                time(12, 00),
                HolidayCalendar(
                    [
                        new_years_eve(
                            start_date=pd.Timestamp("2011-03-07"),
                            days_of_week=weekdays,
                        ),
                        christmas_eve(
                            start_date=pd.Timestamp("2011-03-07"), days_of_week=weekdays
                        ),
                    ]
                ),
            ),
        ]

    @property
    def special_closes_adhoc(self):
        lunar_new_years_eve = (chinese_lunar_new_year_dates - pd.Timedelta(days=1))[
            np.isin(
                chinese_lunar_new_year_dates.weekday,
                [TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY],
            )
            & (chinese_lunar_new_year_dates.year >= 2013)
        ].values

        def selection(arr, start, end):
            predicates = []
            if start is not None:
                predicates.append(start.asm8 <= arr)
            if end is not None:
                predicates.append(arr < end.asm8)

            if not predicates:
                return arr

            return arr[np.all(predicates, axis=0)]

        return [
            (time, pd.DatetimeIndex(selection(lunar_new_years_eve, start, end)))
            for (start, time), (end, _) in toolz.sliding_window(
                2,
                toolz.concatv(self.regular_early_close_times, [(None, None)]),
            )
        ]



import pytest

from exchange_calendars.exchange_calendar_xses import XSESExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase
from .test_utils import T


class TestXSESCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XSESExchangeCalendar

    @pytest.fixture
    def start_bound(self):
        yield T("1986-01-01")

    @pytest.fixture
    def end_bound(self):
        yield T("2025-12-31")

    @pytest.fixture
    def max_session_hours(self):
        # Singapore stock exchange is open from 9am to 5pm
        yield 8

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2017-01-02",
            "2017-01-30",
            "2017-04-14",
            "2017-05-01",
            "2017-05-10",
            "2017-06-26",
            "2017-08-09",
            "2017-09-01",
            "2017-10-18",
            "2017-12-25",
            "2023-12-25",
            "2025-01-01",
            "2025-12-25",
        ]



import pytest

from exchange_calendars.exchange_calendar_xwar import XWARExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXWARCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XWARExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XWAR is open from 9:00AM to 5:00PM.
        yield 8

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2019 +
            "2019-01-01",  # New Year's Day
            "2017-01-06",  # Epiphany
            "2019-04-19",  # Good Friday
            "2019-04-22",  # Easter Monday
            "2019-05-01",  # Labour Day
            "2019-05-03",  # Constitution Day
            "2019-06-20",  # Corpus Christi
            "2019-08-15",  # Armed Forces Day
            "2019-11-01",  # All Saints' Day
            "2019-11-11",  # Independence Day
            "2019-12-24",  # Christmas Eve
            "2019-12-25",  # Christmas Day
            "2019-12-26",  # Boxing Day
            "2019-12-31",  # New Year's Eve
            #
            # New Year's Eve observed as regular holiday only from 2011
            "2012-12-31",  # First year fell on otherwise trading day.
            "2013-12-31",
            #
            # The Epithany observed as regular holiday only from 2011
            "2011-01-06",
            "2012-01-06",
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            "2005-04-08",  # Pope's Funeral.
            "2007-12-31",  # New Year's Eve (adhoc).
            "2008-05-02",  # Exchange Holiday.
            "2009-01-02",  # Exchange Holiday.
            "2013-04-16",  # Exchange Holiday.
            "2018-01-02",  # Exchange Holiday.
            "2018-11-12",  # Independence Holiday.
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # New Year's Eve not a holiday prior to 2011.
            "2008-12-31",
            "2009-12-31",
            "2010-12-31",
            #
            # Epiphany not a holiday prior to 2011.
            "2006-01-06",
            "2009-01-06",
            "2010-01-06",
            #
            # Holidays that fall on a weekend are not made up. Ensure surrounding
            # days are not holidays.
            # New Year's Eve on a Saturday and New Year's Day on a Sunday.
            "2016-12-30",
            "2017-01-02",
            # Epiphany (January 6th) on a Sunday.
            "2019-01-04",
            "2019-01-07",
            # Labour Day (May 1st) on a Sunday.
            "2016-04-29",
            "2016-05-02",
            # Constitution Day (May 3rd) on a Saturday.
            "2014-05-02",
            "2014-05-05",
            # Armed Forces Day (August 15th) on a Saturday.
            "2015-08-14",
            "2015-08-17",
            # All Saints' Day (November 1st) on a Sunday.
            "2015-10-30",
            "2015-11-02",
            # Independence Day (November 11th) on a Saturday.
            "2017-11-10",
            "2017-11-13",
            # Christmas Eve on a Saturday and Christmas on a Sunday. Note that
            # Monday the 26th is Boxing Day, so check the 27th.
            "2016-12-23",
            "2016-12-27",
        ]



import pytest

from exchange_calendars.exchange_calendar_xbru import XBRUExchangeCalendar
from .test_exchange_calendar import EuronextCalendarTestBase


class TestXBRUCalendar(EuronextCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XBRUExchangeCalendar

    @pytest.fixture
    def additional_regular_holidays_sample(self):
        yield [
            # Final observance of these regular holidays
            "2000-06-01",  # Ascension Day
            "2001-06-04",  # Whit Monday
            "2000-07-21",  # Belgium Independence Day
            "2001-12-31",  # New Year's Eve
        ]

    @pytest.fixture
    def additional_non_holidays_sample(self):
        yield [
            # First year when previous regular holiday no longer observed
            "2001-05-24",  # Ascension Day
            "2002-05-20",  # Whit Monday
            "2003-07-21",  # Belgium Independence Day
            "2002-12-31",  # New Year's Eve
        ]



import pandas as pd
import pytest

from exchange_calendars.exchange_calendar_xtae import XTAEExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXTAECalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XTAEExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # Longest session is from 9:59:00 to 17:15:00
        yield 7.25 + (1.0 / 60.0)

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2019
            "2019-03-21",  # Purim
            "2019-04-09",  # Election Day
            "2019-04-25",  # Passover II Eve
            "2019-04-26",  # Passover II
            "2019-05-08",  # Memorial Day
            "2019-05-09",  # Independence Day
            "2019-06-09",  # Pentecost (Shavuot)
            "2019-08-11",  # Fast Day
            "2019-09-17",  # Election Day
            "2019-09-29",  # Jewish New Year Eve
            "2019-09-30",  # Jewish New Year I
            "2019-10-01",  # Jewish New Year II
            "2019-10-08",  # Yom Kiuppur Eve
            "2019-10-09",  # Yom Kippur
            "2019-10-13",  # Feast of Tabernacles (Sukkoth) Eve
            "2019-10-14",  # Feast of Tabernacles
            "2019-10-20",  # Rejoicing of the Law (Simchat Tora) Eve
            "2019-10-21",  # Rejoicing of the Law
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            # Passover interim days
            # 2019
            "2019-04-21",
            "2019-04-22",
            "2019-04-23",
            "2019-04-24",
            # 2020
            "2020-04-12",  # another Sunday (see 2022 comment...)
            # 2022
            # '2022-04-17' is a Sunday. Including here checks holiday early
            # close takes precedence over sunday early close
            "2022-04-17",
            "2022-04-18",
            "2022-04-19",
            "2022-04-20",
            # 2023
            "2023-04-09",
            "2023-04-10",
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=14, minutes=15)

    @pytest.fixture
    def early_closes_weekdays(self):
        return (6,)

    @pytest.fixture
    def early_closes_weekdays_sample(self):
        yield [
            "2022-08-21",  # a sunday of a standard week
        ]

    @pytest.fixture
    def early_closes_weekdays_sample_time(self):
        yield pd.Timedelta(hours=15, minutes=40)

    @pytest.fixture
    def non_early_closes_sample(self):
        yield [
            # check standard week
            # 2022-08-21 is a regular early close sunday
            # check all other days have regular closes
            "2022-08-17",
            "2022-08-18",
            "2022-08-22",
            "2022-08-23",
        ]

    @pytest.fixture
    def non_early_closes_sample_time(self):
        yield pd.Timedelta(hours=17, minutes=15)



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_cmes import CMESExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestCMESCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class", params=["left", "right"])
    def all_calendars_with_answers(self, request, calendars, answers):
        yield (calendars[request.param], answers[request.param])

    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield CMESExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 24

    @pytest.fixture
    def regular_holidays_sample(self):
        # good friday, christmas, new years
        yield ["2016-03-25", "2016-12-26", "2016-01-02"]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2016-01-18",  # mlk day
            "2016-02-15",  # presidents
            "2016-05-30",  # mem day
            "2016-07-04",  # july 4
            "2016-09-05",  # labor day
            "2016-11-24",  # thanksgiving
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(12, "h")



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xkar import XKARExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXKARCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XKARExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 5.967

    @pytest.fixture
    def regular_holidays_sample(self):
        # iqbal day observed until 2014 (last year of observance 2013)
        yield [f"{year}-11-09" for year in range(2002, 2014)]

    @pytest.fixture
    def non_holidays_sample(self):
        # iqbal day not observed from 2014
        yield [
            f"{year}-11-09"
            for year in range(2014, 2019)
            if pd.Timestamp(f"{year}-11-09").weekday() not in (5, 6)
        ]

    @pytest.mark.parametrize(
        "year, holidays",
        [
            # https://www.psx.com.pk/psx/exchange/general/calendar-holidays
            (
                2019,
                [
                    "2019-02-05",  # Kashmir Day
                    "2019-03-23",  # Pakistan Day
                    "2019-05-01",  # Labour Day
                    "2019-05-31",  # Juma-Tul-Wida
                    "2019-06-04",  # [NOTE: not included on the website]
                    "2019-06-05",  # Eid-ul-Fitr
                    "2019-06-06",  # Eid-ul-Fitr
                    "2019-06-07",  # Eid-ul-Fitr
                    "2019-08-12",  # Eid-ul-Azha
                    "2019-08-13",  # Eid-ul-Azha
                    "2019-08-14",  # Independence Day
                    "2019-08-15",  # Eid-ul-Azha
                    "2019-09-09",  # Muharram (Ashura)
                    "2019-09-10",  # Muharram (Ashura)
                    "2019-11-10",  # Eid Milad-un-Nabi (SAW)
                    "2019-12-25",  # Birthday of Quaid-e-Azam & Christmas
                ],
            ),
        ],
    )
    def test_holidays_in_year(self, default_calendar, year, holidays):
        cal = default_calendar
        days = pd.date_range(start=f"{year}-01-01", end=f"{year}-12-31", freq="B")
        days = days.strftime("%Y-%m-%d")

        for holiday in holidays:
            # Sanity check
            assert holiday in days or pd.Timestamp(holiday).weekday() in (5, 6)

        for day in days:
            if day in holidays:
                assert not cal.is_session(day)
            else:
                assert cal.is_session(day)



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xsgo import XSGOExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase
from .test_utils import T


class TestXSGOCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XSGOExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # XSGO's longest session occurs when open from 9:30AM to 5:00PM.
        yield 7.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2019
            "2019-01-01",  # New Year's Day
            "2019-04-19",  # Good Friday
            "2019-05-01",  # Labour Day
            "2019-05-21",  # Navy Day
            "2019-07-16",  # Our Lady of Mount Carmel Day
            "2019-08-15",  # Assumption Day
            "2019-09-18",  # Independence Day
            "2019-09-19",  # Army Day
            "2019-09-20",  # Public Holiday
            "2019-10-31",  # Evangelical Church Day
            "2019-11-01",  # All Saints' Day
            "2017-12-08",  # Immaculate Conception
            "2019-12-25",  # Christmas Day
            "2019-12-31",  # Bank Holiday
            #
            # Saint Peter and Saint Paul Day follows an unusual rule whereby if June
            # 29th falls on a Saturday, Sunday or Monday then the holiday is
            # acknowledged on that day (Sat/Sun observances are not made up later).
            # Otherwise it is observed on the closest Monday to the 29th.
            "2018-07-02",  # 29th a Friday so next Monday is a holiday.
            "2017-06-26",  # 29th a Thursday so previous Monday is a holiday.
            "2016-06-27",  # 29th a Wednesday so previous Monday is a holiday.
            "2015-06-29",  # 29th a Monday so that Monday is a holiday.
            "2010-06-28",  # 29th a Tuesday so previous Monday is a holiday.
            #
            # Dia de la Raza (also known as Columbus Day) follows same rule as Saint
            # Peter and Saint Paul Day described above, albeit centered on October 12th.
            "2018-10-15",  # 12th a Friday so next Monday is a holiday.
            "2017-10-09",  # 12th a Thursday so previous Monday is a holiday.
            "2016-10-10",  # 12th a Wednesday so previous Monday is a holiday.
            "2015-10-12",  # 12th a Monday so that Monday is a holiday.
            "2010-10-11",  # 12th a Tuesday so previous Monday is a holiday.
            #
            # Independence Day and Army Day fall back-to-back on September 18th and
            # 19th. If they happen to fall on a Tue/Wed then the prior Monday is
            # deemed a Public Holiday. If they happen to fall on a Wed/Thu then the
            # following Friday is deemed a Public Holiday.
            "2019-09-20",  # fall on Wed/Thu so following Friday 20th is also a holiday.
            "2018-09-17",  # fall on Tue/Wed so prior Monday 17th is also a holiday.
            #
            # Evangelical Church Day (also known as Halloween) adheres to the
            # following rule: If October 31st falls on a Tuesday, it is observed the
            # preceding Friday. If it falls on a Wednesday, it is observed the next
            # Friday. If it falls on Thu, Fri, Sat, Sun, or Mon then the holiday is
            # acknowledged that day.
            "2019-10-31",  # 31st is a Thursday, so that Thursday is a holiday.
            "2018-11-02",  # 31st is a Wednesday, so following Friday is a holiday.
            "2017-10-27",  # 31st is a Tuesday, so previous Friday is a holiday.
            "2016-10-31",  # 31st is a Monday, so that Monday is a holiday.
            "2014-10-31",  # 31st is a Friday, so that Friday is a holiday.
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            "2010-09-17",  # Bicentennial Celebration.
            "2010-09-20",  # Bicentennial Celebration.
            # For whatever reason New Year's Day, which was a Sunday, was
            # observed on Monday this one year.
            "2017-01-02",
            "2017-04-19",  # Census Day.
            "2018-01-16",  # Pope Visit.
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend and are not made up. Ensure surrounding
            # days are not holidays.
            # Bank Holiday on Saturday, December 31st and  New Year's Day on
            # Sunday, January 1st.
            "2011-12-30",
            "2012-01-02",
            # Labour Day on Sunday, May 1st.
            "2016-04-29",
            "2016-05-02",
            # Navy Day on Sunday, May 21st.
            "2017-05-19",
            "2017-05-22",
            # Saint Peter and Saint Paul Day on Saturday, June 29th.
            "2019-06-28",
            "2019-07-01",
            # Our Lady of Mount Carmel Day on Sunday, July 16th.
            "2017-07-14",
            "2017-07-17",
            # Assumption Day on Saturday, August 15th.
            "2015-08-14",
            "2015-08-17",
            # In 2004 Independence Day and Army Day fell on a Saturday and
            # Sunday, so the surrounding Friday and Monday should both be
            # trading days.
            "2004-09-17",
            "2004-09-20",
            # Dia de la Raza on Saturday, October 12th.
            "2019-10-11",
            "2019-10-14",
            # Evangelical Church Day (Halloween) and All Saints' Day fall on a
            # Saturday and Sunday, so Friday the 30th and Monday the 2nd should
            # both be trading days.
            "2015-10-30",
            "2015-11-02",
            # Immaculate Conception on Sunday, December 8th.
            "2019-12-06",
            "2019-12-09",
            # Christmas on a Sunday.
            "2016-12-23",
            "2016-12-26",
            #
            # Saint Peter and Saint Paul Day follows an unusual rule whereby if June
            # 29th falls on a Saturday, Sunday or Monday then the holiday is
            # acknowledged on that day (Sat/Sun observances are not made up later).
            # Otherwise it is observed on the closest Monday to the 29th.
            "2014-06-30",  # 29th is a Sunday, so closest Monday is not a holiday
            "2019-07-01",  # 29th a Saturday, so closest Monday a trading day
            #
            # # Dia de la Raza (also known as Columbus Day) follows same rule as Saint
            # Peter and Saint Paul Day described above, albeit centered on October 12th.
            "2014-10-13",  # 12th is a Sunday, so closest Monday is not a holiday
            "2019-10-14",  # 12th a Saturday, so closest Monday a trading day
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            "2019-04-18",  # Maundy Thursday
            "2019-09-17",  # Day before Independence Day
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=13, minutes=30)

    # Calendar-specific tests

    def test_additional_early_closes_sample(self, default_calendar):
        # additional test necessray as early close time here (12:30 local)
        # is different to that for sessions of `early_closes_sample` (13:30 local).
        cal = default_calendar
        early_closes = [
            "2019-12-24",  # Christmas Eve
            "2019-12-30",  # Day before Bank Holiday
        ]

        for date in early_closes:
            assert T(date) in cal.early_closes

        tz = default_calendar.tz
        early_closes_time = pd.Timedelta(hours=12, minutes=30)
        offset = pd.Timedelta(cal.close_offset, "D") + early_closes_time
        for date in early_closes:
            early_close = cal.closes[date].tz_convert(tz)
            expected = pd.Timestamp(date, tz=tz) + offset
            assert early_close == expected

    def test_close_time_change(self, default_calendar):
        """Test twice annual change to close time."""
        # Evert March close time changes from 5:00PM to 4:00PM and every
        # November changes back from 4:00PM to 5:00PM (all local times).
        dates_closes = (
            ("2019-02-28", "2019-02-28 17:00"),
            ("2019-03-01", "2019-03-01 16:00"),
            ("2019-10-30", "2019-10-30 16:00"),
            ("2019-11-04", "2019-11-04 17:00"),
        )
        cal = default_calendar
        for date, close in dates_closes:
            assert cal.closes[date] == pd.Timestamp(close, tz=cal.tz)



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xbue import XBUEExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXBUECalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XBUEExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XBUE is open from 11:00AM to 5:00PM
        yield 6

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2019
            "2019-01-01",  # New Year's Day
            "2019-03-04",  # Carnival Monday
            "2019-03-05",  # Carnival Tuesday
            "2017-03-24",  # Truth and Justice Mem Day
            "2019-04-02",  # Malvinas Day
            "2019-04-13",  # Maundy Thursday
            "2019-04-14",  # Good Friday
            "2019-05-01",  # Labour Day
            "2018-05-25",  # May Day Revolution
            "2019-06-17",  # Martin Miguel de-Guemes Day
            "2019-06-20",  # National Flag Day
            "2019-07-09",  # Independence Day
            "2019-08-19",  # San Martin's Day
            "2019-10-14",  # Cultural Diversity Day
            "2019-11-18",  # Day of Natl Sovereignty
            "2017-12-08",  # Immaculate Conception
            "2019-12-25",  # Christmas Day
            #
            # Day of Respect for Cultural Diversity follows a "nearest Monday"
            # rule. When Oct 12 falls on a Tuesday or Wednesday the holiday is
            # observed on the previous Monday, and when it falls on any other
            # non-Monday it is observed on the following Monday.  This means
            # the holiday will be observed on the following dates (Mondays).
            "2019-10-14",  # Falls on Saturday
            "2018-10-15",  # Falls on Friday
            "2017-10-16",  # Falls on Thursday
            "2016-10-10",  # Falls on Wednesday
            "2015-10-12",  # Falls on Monday
            "2014-10-13",  # Falls on Sunday
            "2013-10-14",  # Falls on Saturday
            "2010-10-11",  # Falls on Tuesday
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend are not made up. Ensure surrounding
            # days are not holidays.
            # New Year's Day on Sunday, Jan 1st.
            "2016-12-30",
            "2017-01-02",
            # Truth and Justice Memorial Day on Sunday, Mar 24th.
            "2019-03-22",
            "2019-03-25",
            # Malvinas Day on Saturday, Apr 2nd.
            "2016-04-01",
            "2016-04-04",
            # Labour Day on Sunday, May 1st.
            "2016-04-29",
            "2016-05-02",
            # May Day Revolution on Saturday, May 25th.
            "2019-05-24",
            "2019-05-27",
            # Martin Miguel de-Guemes Day on Sunday, Jun 17th.
            "2018-06-15",
            "2018-06-18",
            # National Flag Day on Saturday, Jun 20th.
            "2015-06-19",
            "2015-06-22",
            # Independence Day on Sunday, Jul 9th.
            "2017-07-07",
            "2017-07-10",
            # Bank Holiday on Sunday, Nov 6th.
            "2016-11-04",
            "2016-11-07",
            # Immaculate Conception on Sunday, Dec 8th.
            "2019-12-06",
            "2019-12-09",
            # Christmas on Sunday
            "2016-12-23",
            "2016-12-26",
        ]

    @pytest.fixture
    def early_closes_sample(self):
        # Christmas Eve, New Year's Eve
        yield ["2019-12-24", "2019-12-31"]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(13, "h")



from datetime import time
from zoneinfo import ZoneInfo

from pandas.tseries.holiday import EasterMonday, GoodFriday

from .exchange_calendar import WEEKDAYS, HolidayCalendar, ExchangeCalendar

from .common_holidays import (
    new_years_eve,
    new_years_day,
    european_labour_day,
    christmas,
    christmas_eve,
    boxing_day,
)

NewYearsEve = new_years_eve(days_of_week=WEEKDAYS)
NewYearsDay = new_years_day()
LabourDay = european_labour_day()
Christmas = christmas()
ChristmasEve = christmas_eve(days_of_week=WEEKDAYS)
BoxingDay = boxing_day()


class XLUXExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Luxembourg Stock Exchange (XLUX).

    Open Time: 9:00 AM, CET
    Close Time: 5:40 PM, CET

    Regularly-Observed Holidays:
    - New Years Day
    - Good Friday
    - Easter Monday
    - Labour Day
    - Christmas Day
    - Boxing Day

    Early Closes:
    - Christmas Eve
    - New Year's Eve
    """

    # Source: https://www.luxse.com/trading/opening-hours-and-closing-days

    name = "XLUX"  # Luxembourg Stock Exchange
    tz = ZoneInfo("Europe/Luxembourg")
    open_times = ((None, time(9, 0)),)
    close_times = ((None, time(17, 40)),)
    regular_early_close = time(14, 5)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                GoodFriday,
                EasterMonday,
                LabourDay,
                Christmas,
                BoxingDay,
            ]
        )

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [
                        ChristmasEve,
                        NewYearsEve,
                    ]
                ),
            )
        ]
